# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
STAC utils
"""
import os
from datetime import datetime
from pprint import pformat
from typing import Any

import geopandas as gpd
from sertit.vectors import WGS84
from shapely.geometry import mapping

from eoreader.stac._stac_keywords import DESCRIPTION, GSD, TITLE


def fill_common_mtd(asset: Any, prod, **kwargs) -> None:
    """
    Fill common metadata of STAC assets or items

    Args:
        asset (pystac.Asset, pystac.Item): Asset or item
        prod (Product): EOReader product
        **kwargs: Additional arguments
    """
    # Basics
    asset.common_metadata.title = kwargs.get(TITLE, "No title available")
    asset.common_metadata.description = kwargs.get(DESCRIPTION, "")
    if len(asset.common_metadata.description) < 1:
        asset.common_metadata.description = "No description available"

    # Date and Time
    asset.common_metadata.created = datetime.utcnow()
    asset.common_metadata.updated = datetime.utcnow()

    # Licensing
    # asset.common_metadata.license = None  # Collection level if possible

    # Provider
    # asset.common_metadata.providers = None  # Collection level if possible

    # Date and Time Range
    asset.common_metadata.start_datetime = prod.datetime
    asset.common_metadata.end_datetime = prod.datetime  # TODO

    # Instrument
    asset.common_metadata.platform = None  # TODO
    asset.common_metadata.instruments = [
        prod.instrument if isinstance(prod.instrument, str) else prod.instrument.value
    ]
    asset.common_metadata.constellation = prod.constellation.value.lower()
    asset.common_metadata.mission = None
    asset.common_metadata.gsd = kwargs.get(GSD)


def gdf_to_geometry(gdf: gpd.GeoDataFrame) -> dict:
    """
    Convert a geodataframe to a STAC geometry

    Args:
        gdf (gpd.GeoDataFrame): Geodataframe to convert to geometry

    Returns:
        dict: STAC Geometry
    """
    return mapping(gdf.geometry.iat[0])


def gdf_to_bbox(gdf: gpd.GeoDataFrame) -> list:
    """
    Convert a geodataframe to a STAC bbox

    Args:
        gdf (gpd.GeoDataFrame): Geodataframe to convert to bbox

    Returns:
        dict: STAC bbox
    """
    # Convert from numpy dtype (which are not JSON serializable) to standard dtype
    return [float(val) for val in gdf.bounds.values[0]]


def gdf_to_centroid(gdf: gpd.GeoDataFrame) -> dict:
    """
    Convert a geodataframe to a STAC centroid

    Args:
        gdf (gpd.GeoDataFrame): Geodataframe to convert to centroid

    Returns:
        dict: STAC centroid
    """
    centroid = gdf.centroid.to_crs(WGS84).iat[0]

    # Convert from numpy dtype (which are not JSON serializable) to standard dtype
    return {"lat": float(centroid.y), "lon": float(centroid.x)}


def repr_multiline_str(to_str: Any, nof_tabs: int) -> str:
    """
    Format multiline string repr of a list, dict... by adding tabs at the beginning of every new line.

    Begins with a new line.

    Args:
        to_str (Any): List, dict... to convert to tabbed string
        nof_tabs (int): Number of tabs to put at the beginning of the lines

    Returns:

    """
    nof_tabs_str = "\t" * nof_tabs
    return f"\n{nof_tabs_str}" + f"{nof_tabs_str}".join(
        pformat(to_str).splitlines(True)
    )


def get_media_type(band_path: str) -> Any:
    """
    Get media type

    Args:
        band_path (str): Band path

    Returns:
        Any: Media Type

    """
    import pystac

    suffix = os.path.splitext(band_path)[-1]
    if suffix.lower() in [".tiff", ".tif"]:
        # Manage COGs
        media_type = pystac.MediaType.GEOTIFF
    elif suffix.lower() == ".jp2":
        media_type = pystac.MediaType.JPEG2000
    elif suffix.lower() == ".xml":
        media_type = pystac.MediaType.XML
    elif suffix.lower() == ".png":
        media_type = pystac.MediaType.PNG
    elif suffix.lower() in [".jpeg", ".jpg"]:
        media_type = pystac.MediaType.JPEG
    elif suffix.lower() == ".til":
        media_type = None  # Not existing
    elif suffix.lower() == ".nc":
        media_type = None  # Not existing
    else:
        media_type = None  # Not recognized

    return media_type


def to_float(val: Any) -> float:
    """
    Convert from numpy dtype (which are not JSON serializable) to standard dtype

    Args:
        val (Any): Value in numpy dtype to be converted to float

    Returns:
        float: Value converted

    """
    try:
        # Convert from numpy dtype (which are not JSON serializable) to standard dtype
        return float(val)
    except TypeError:
        return val



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SPOT-6/7 products.
See `here <https://earth.esa.int/eogateway/documents/20142/37627/SPOT-6-7-imagery-user-guide.pdf>`_
for more information.
"""
import logging

from eoreader import EOREADER_NAME
from eoreader.bands import BLUE, GREEN, NIR, PAN, RED, SpectralBand
from eoreader.products import DimapV2Product
from eoreader.reader import Constellation
from eoreader.stac import GSD, ID, NAME, WV_MAX, WV_MIN

LOGGER = logging.getLogger(EOREADER_NAME)


class Spot67Product(DimapV2Product):
    """
    Class of SPOT-6/7 products.
    See `here <https://earth.esa.int/eogateway/documents/20142/37627/SPOT-6-7-imagery-user-guide.pdf>`_
    for more information.
    """

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._pan_res = 1.5
        self._ms_res = 6.0
        self._altitude = 697000

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _get_constellation(self) -> Constellation:
        """Getter of the constellation"""
        constellation_id = self.split_name[0]
        return getattr(Constellation, constellation_id)

    def _map_bands(self) -> None:
        """Set products type"""
        # Create spectral bands
        # https://www.intelligence-airbusds.com/automne/api/docs/v1.0/document/download/ZG9jdXRoZXF1ZS1kb2N1bWVudC01NTMyMw==/ZG9jdXRoZXF1ZS1maWxlLTU1MzIy/spot-brochure-2019.pdf
        pan = SpectralBand(
            eoreader_name=PAN,
            **{NAME: "PAN", ID: 1, GSD: 1.5, WV_MIN: 450, WV_MAX: 745}
        )

        blue = SpectralBand(
            eoreader_name=BLUE,
            **{NAME: "BLUE", ID: 1, GSD: 6, WV_MIN: 450, WV_MAX: 520}
        )

        green = SpectralBand(
            eoreader_name=GREEN,
            **{NAME: "GREEN", ID: 2, GSD: 6, WV_MIN: 530, WV_MAX: 590}
        )

        red = SpectralBand(
            eoreader_name=RED, **{NAME: "RED", ID: 3, GSD: 6, WV_MIN: 625, WV_MAX: 695}
        )

        nir = SpectralBand(
            eoreader_name=NIR, **{NAME: "NIR", ID: 4, GSD: 6, WV_MIN: 760, WV_MAX: 890}
        )
        self._map_bands_core(blue=blue, green=green, red=red, nir=nir, pan=pan)

    def _set_instrument(self) -> None:
        """
        Set instrument

        SPOT-6/7: https://earth.esa.int/eogateway/missions/spot-6
        """
        # NAOMI: New Astrosat Optical Modular Instrument
        self.instrument = "NAOMI"



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SPOT-4/5 products.
See `here <http://www.engesat.com.br/wp-content/uploads/S5-ST-73-1-CN_2_9-Spec-Format-Produits-SPOT.pdf>`_
for more information.
"""
import logging
from datetime import timedelta
from enum import unique

import numpy as np
import xarray as xr
from lxml import etree
from rasterio import crs as riocrs
from sertit import path
from sertit.misc import ListEnum
from sertit.types import AnyPathType

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import (
    GREEN,
    NARROW_NIR,
    NIR,
    PAN,
    RED,
    SWIR_1,
    BandNames,
    SpectralBand,
)
from eoreader.exceptions import InvalidProductError
from eoreader.products import DimapV1Product
from eoreader.products.optical.optical_product import RawUnits
from eoreader.reader import Constellation
from eoreader.stac import GSD, ID, NAME, WV_MAX, WV_MIN

LOGGER = logging.getLogger(EOREADER_NAME)

_DIMAP_BAND_MTD = {
    PAN: "PAN",
    GREEN: "XS1",
    RED: "XS2",
    NIR: "XS3",
    NARROW_NIR: "XS3",
    SWIR_1: "XS4",
}

# https://spot.cnes.fr/sites/default/files/migration/smsc/spot/calibration_synthesis_SPOT1245_ed1.pdf (3.3)
# HRVIR 1, 2
_SPOT4_E0 = {
    PAN: [1570.2, 1589],
    GREEN: [1842.9, 1850.9],
    RED: [1570.2, 1589],
    NIR: [1052.1, 1054.8],
    NARROW_NIR: [1052.1, 1054.8],
    SWIR_1: [235.84, 241.93],
}

# HRG 1, 2
_SPOT5_E0 = {
    PAN: [1764.2, 1775],
    GREEN: [1859.8, 1859.8],
    RED: [1575.3, 1577.6],
    NIR: [1043.9, 1048.2],
    NARROW_NIR: [1043.9, 1048.2],
    SWIR_1: [238.87, 237.78],
}
"""
The values of the normalized solar irradiance have been computed using WMO (World Meteorogical Organization) spectral solar irradiance.
"""


@unique
class Spot4BandCombination(ListEnum):
    """
    Band combination for SPOT4 data
    See `this <https://www.intelligence-airbusds.com/files/pmedia/public/r451_9_resolutionspectralmodes_uk_sept2010.pdf>`_ for more information.
    """

    M = "M"
    """
    "M" for Spot 4 PAN product (10 m)
    """

    X = "X"
    """
    "X" for Spot 4 multispectral product (3 bands, without SWIR) (20 m)
    """

    I = "I"  # noqa
    """
    "I" for Spot 4 multispectral product (4 bands, with SWIR) (20 m)
    """

    MX = "M+X"
    """
    "M+X" for Spot 4 merge product (3 bands, without SWIR) (10 m)
    """

    MI = "M+I"
    """
    "M+I" for Spot 4 merge product (4 bands, with SWIR) (10 m)
    """


@unique
class Spot5BandCombination(ListEnum):
    """
    Band combination for SPOT4/5 data
    See `this <https://www.intelligence-airbusds.com/files/pmedia/public/r451_9_resolutionspectralmodes_uk_sept2010.pdf>`_ for more information.
    """

    T = "T"
    """
    "T" for supermode Spot data (2.5 m)
    """

    HM = "HM"
    """
    "HM" for Spot-5 PAN data (5 m)
    """

    X = "X"
    """
    "X" for 3 bands extracted from 4 bands (10 m)
    """

    J = "J"
    """
    "J" for Spot 5 multipectral product (4 bands, with SWIR) (10 m)
    """

    HMX = "HM+X"
    """
    "HM+X" for Spot 5 merge product (3 bands, without SWIR) (5 m)
    """

    TX = "T+X"
    """
    "T+X" for Spot 5 supermode and multipectral merge product (3 bands, without SWIR) (2.5 m)
    """


@unique
class Spot45ProductType(ListEnum):
    """
    Product Type for SPOT4/5 data
    See `here <http://www.engesat.com.br/wp-content/uploads/S5-ST-73-1-CN_2_9-Spec-Format-Produits-SPOT.pdf>`_ for more information.
    """

    L0 = "0"
    """
    Level-0
    """

    L1A = "1A"
    """
    Level-1A
    """

    L1B = "1B"
    """
    Level-1B
    """

    L2A = "2A"
    """
    Level-2A
    """


class Spot45Product(DimapV1Product):
    """
    Class of SPOT4/5 products.
    See `here <http://www.engesat.com.br/wp-content/uploads/S5-ST-73-1-CN_2_9-Spec-Format-Produits-SPOT.pdf>`_
    for more information.
    """

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        root, _ = self.read_mtd()
        mission_idx = int(root.findtext(".//MISSION_INDEX"))
        if mission_idx == 4:
            self._supermode_res = None
            self._pan_res = 10.0
            self._ms_res = 20.0
            self.constellation = Constellation.SPOT4
        elif mission_idx == 5:
            self._supermode_res = 2.5
            self._pan_res = 5.0
            self._ms_res = 10.0
            self.constellation = Constellation.SPOT5
        else:
            raise InvalidProductError("Mission index should be 4 or 5.")

        self.needs_extraction = False
        self._use_filename = True
        self._proj_prod_type = [Spot45ProductType.L0]

        # Raw units
        rad_proc = root.findtext(".//RADIOMETRIC_PROCESSING").upper()

        if rad_proc == "REFLECTANCE":
            self._raw_units = RawUnits.REFL
        elif rad_proc in "BASIC":
            self._raw_units = RawUnits.DN
        else:
            self._raw_units = RawUnits.NONE

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _set_band_combi(self) -> None:
        """
        Set Band combination
        """
        root, _ = self.read_mtd()
        band_combi = root.findtext(".//SPECTRAL_PROCESSING")
        if self.constellation == Constellation.SPOT4:
            self.band_combi = Spot4BandCombination.from_value(band_combi)
        else:
            self.band_combi = Spot5BandCombination.from_value(band_combi)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        if self.band_combi is None:
            self._set_band_combi()

        # Post init done by the super class
        super()._post_init(**kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        # Not Pansharpened images
        if self.band_combi in [
            Spot4BandCombination.X,
            Spot4BandCombination.I,
            Spot5BandCombination.X,
            Spot5BandCombination.J,
        ]:
            self.pixel_size = self._ms_res
        # Pansharpened images
        elif self.band_combi in [
            Spot4BandCombination.M,
            Spot4BandCombination.MX,
            Spot4BandCombination.MI,
            Spot5BandCombination.HM,
            Spot5BandCombination.HMX,
        ]:
            self.pixel_size = self._pan_res
        # Supermode images
        else:
            self.pixel_size = self._supermode_res

    def _set_instrument(self) -> None:
        """
        Set instrument

        SPOT-4: https://space.oscar.wmo.int/instruments/view/hrvir
        SPOT-5: https://space.oscar.wmo.int/instruments/view/hrg
        """
        if self.constellation == Constellation.SPOT4:
            self.instrument = "HRVIR"
        else:
            self.instrument = "HRG"

    def _set_product_type(self) -> None:
        """
        Set products type

        See Vision-1_web_201906.pdf for more information.
        """
        if self.product_type is None:
            # Get MTD XML file
            root, _ = self.read_mtd()
            proc_lvl = root.findtext(".//SCENE_PROCESSING_LEVEL")
            self.product_type = Spot45ProductType.from_value(proc_lvl)

            # Manage not orthorectified product
            if self.product_type == Spot45ProductType.L0:
                self.is_ortho = False

    def _map_bands(self) -> None:
        """
        Map bands
        """
        if self.constellation == Constellation.SPOT4:
            # Create spectral bands
            pan = SpectralBand(
                eoreader_name=PAN,
                **{NAME: "PAN", ID: 1, GSD: self._pan_res, WV_MIN: 610, WV_MAX: 680},
            )

            green = SpectralBand(
                eoreader_name=GREEN,
                **{NAME: "GREEN", ID: 3, GSD: self._ms_res, WV_MIN: 500, WV_MAX: 590},
            )

            red = SpectralBand(
                eoreader_name=RED,
                **{NAME: "RED", ID: 2, GSD: self._ms_res, WV_MIN: 610, WV_MAX: 680},
            )

            nir = SpectralBand(
                eoreader_name=NIR,
                **{NAME: "NIR", ID: 1, GSD: self._ms_res, WV_MIN: 790, WV_MAX: 890},
            )

            swir1 = SpectralBand(
                eoreader_name=SWIR_1,
                **{
                    NAME: "SWIR_1",
                    ID: 4,
                    GSD: self._ms_res,
                    WV_MIN: 1580,
                    WV_MAX: 1750,
                },
            )
        else:
            # Create spectral bands
            pan = SpectralBand(
                eoreader_name=PAN,
                **{NAME: "PAN", ID: 1, GSD: self._pan_res, WV_MIN: 490, WV_MAX: 690},
            )

            green = SpectralBand(
                eoreader_name=GREEN,
                **{NAME: "GREEN", ID: 3, GSD: self._ms_res, WV_MIN: 490, WV_MAX: 610},
            )

            red = SpectralBand(
                eoreader_name=RED,
                **{NAME: "RED", ID: 2, GSD: self._ms_res, WV_MIN: 610, WV_MAX: 680},
            )

            nir = SpectralBand(
                eoreader_name=NIR,
                **{NAME: "NIR", ID: 1, GSD: self._ms_res, WV_MIN: 780, WV_MAX: 890},
            )

            swir1 = SpectralBand(
                eoreader_name=SWIR_1,
                **{NAME: "SWIR_1", ID: 4, GSD: 20.0, WV_MIN: 1580, WV_MAX: 1750},
            )

        # Manage bands of the product
        if self.band_combi in [Spot4BandCombination.M, Spot5BandCombination.HM]:
            self.bands.map_bands({PAN: pan})
        elif self.band_combi in [
            Spot4BandCombination.X,
            Spot5BandCombination.X,
        ]:
            self.bands.map_bands(
                {
                    GREEN: green,
                    RED: red,
                    NIR: nir,
                    NARROW_NIR: nir,
                }
            )
        elif self.band_combi in [
            Spot4BandCombination.MX,
            Spot5BandCombination.HMX,
        ]:
            self.bands.map_bands(
                {
                    GREEN: green.update(gsd=self._pan_res),
                    RED: red.update(gsd=self._pan_res),
                    NIR: nir.update(gsd=self._pan_res),
                    NARROW_NIR: nir.update(gsd=self._pan_res),
                }
            )
        elif self.band_combi == Spot5BandCombination.T:
            self.bands.map_bands({PAN: pan.update(gsd=self._supermode_res)})
        elif self.band_combi == Spot5BandCombination.TX:
            self.bands.map_bands(
                {
                    GREEN: green.update(gsd=self._supermode_res),
                    RED: red.update(gsd=self._supermode_res),
                    NIR: nir.update(gsd=self._supermode_res),
                    NARROW_NIR: nir.update(gsd=self._supermode_res),
                }
            )
        elif self.band_combi in [
            Spot4BandCombination.I,
            Spot5BandCombination.J,
        ]:
            self.bands.map_bands(
                {
                    GREEN: green,
                    RED: red,
                    NIR: nir,
                    NARROW_NIR: nir,
                    SWIR_1: swir1,
                }
            )
        elif self.band_combi == Spot4BandCombination.MI:
            self.bands.map_bands(
                {
                    GREEN: green.update(gsd=self._pan_res),
                    RED: red.update(gsd=self._pan_res),
                    NIR: nir.update(gsd=self._pan_res),
                    NARROW_NIR: nir.update(gsd=self._pan_res),
                    SWIR_1: swir1.update(gsd=self._pan_res),
                }
            )
        else:
            raise InvalidProductError(
                f"Unusual band combination: {self.band_combi.name}"
            )

    def _get_raw_crs(self) -> riocrs.CRS:
        """
        Get raw CRS of the tile

        Returns:
            rasterio.crs.CRS: CRS object
        """
        # Open metadata
        root, _ = self.read_mtd()

        # Get CRS
        crs_name = root.findtext(".//HORIZONTAL_CS_CODE")
        if not crs_name:
            raise InvalidProductError(
                "Cannot find the CRS name (from GEOGRAPHIC_CS_CODE or HORIZONTAL_CS_CODE) type in the metadata file"
            )

        return riocrs.CRS.from_string(crs_name)

    def _get_constellation(self) -> Constellation:
        """Getter of the constellation"""
        if self.split_name[0] == "SP04":
            const = Constellation.SPOT4
        elif self.split_name[0] == "SP05":
            const = Constellation.SPOT5
        else:
            raise InvalidProductError(
                f"Invalid name: {self.name}, should start with SP04 or SP05."
            )

        return const

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Mission index
        mission_idx = int(root.findtext(".//MISSION_INDEX"))

        # Instrument
        instrument = "HIR" if self.constellation == Constellation.SPOT4 else "HRG"

        # Product type
        self._set_band_combi()  # Not set yet
        band_combi = f"{self.band_combi.name:_<4}"

        # Datetimes
        dt = self.get_datetime(as_datetime=True)
        start_dt = (dt - timedelta(seconds=4)).strftime(DATETIME_FMT)
        end_dt = (dt + timedelta(seconds=4)).strftime(DATETIME_FMT)

        # Unknown data
        digit = 3  # TODO: what's this ?
        suffix = "TOU_1234_eord"

        # Create name
        name = f"SP0{mission_idx}_{instrument}_{band_combi}_{digit}_{start_dt}_{end_dt}_{suffix}"

        return name

    @cache
    def get_mean_viewing_angles(self) -> (float, float, float):
        """
        Get Mean Viewing angles (azimuth, off-nadir and incidence angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_viewing_angles()

        Returns:
            (float, float, float): Mean azimuth, off-nadir and incidence angles
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open incidence and off-nadir angle
        az = None
        try:
            incidence_angle = abs(float(root.findtext(".//INCIDENCE_ANGLE")))
        except TypeError:
            raise InvalidProductError(
                "INCIDENCE_ANGLE or VIEWING_ANGLE not found in metadata!"
            )
        if self.constellation == Constellation.SPOT5:
            try:
                off_nadir = abs(float(root.findtext(".//VIEWING_ANGLE")))
            except TypeError:
                raise InvalidProductError("VIEWING_ANGLE not found in metadata!")
        else:
            # See: https://earth.esa.int/eogateway/missions/spot-4
            orbit_height = 832000
            earth_radius = 6378137
            orbit_coeff = (earth_radius + orbit_height) / earth_radius
            off_nadir = np.rad2deg(
                np.arcsin(np.sin(np.deg2rad(90 - incidence_angle)) / orbit_coeff)
            )

        return az, off_nadir, incidence_angle

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        if self._raw_units == RawUnits.REFL:
            # Compute the correct radiometry of the band
            original_dtype = band_arr.encoding["dtype"]
            if original_dtype == "uint16":
                band_arr /= 10000.0
        elif self._raw_units == RawUnits.DN:
            # Convert DN into radiance
            band_arr = self._dn_to_toa_rad(band_arr, band)

            # Get the solar irradiance value of raw radiometric Band (in watt/m2/micron)
            root, _ = self.read_mtd()
            instrument_idx = int(root.findtext(".//INSTRUMENT_INDEX")) - 1

            if self.constellation == Constellation.SPOT4:
                e0 = _SPOT4_E0[band][instrument_idx]
            else:
                e0 = _SPOT5_E0[band][instrument_idx]

            # Convert radiance into reflectance
            band_arr = self._toa_rad_to_toa_refl(band_arr, band, e0=e0)

        else:
            LOGGER.warning(
                "The spectral properties of a SEAMLESS radiometric processed image "
                "cannot be retrieved since the initial images have undergone "
                "several radiometric adjustments for aesthetic rendering."
                "Returned as is."
            )

        # To float32
        if band_arr.dtype != np.float32:
            band_arr = band_arr.astype(np.float32)

        return band_arr

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces as a dict
        """
        mtd_from_path = "METADATA.DIM"
        mtd_archived = r"METADATA\.DIM"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def _get_tile_path(self) -> AnyPathType:
        """
        Get the DIMAP filepath

        Returns:
            AnyPathType: DIMAP filepath
        """
        return self._get_path("IMAGERY", "TIF")

    def _dn_to_toa_rad(self, dn_arr: xr.DataArray, band: BandNames) -> xr.DataArray:
        """
        Compute DN to TOA radiance

        Args:
            dn_arr (xr.DataArray): DN array
            band (BandNames): Band

        Returns:
            xr.DataArray: TOA Radiance array
        """
        band_mtd_str = _DIMAP_BAND_MTD[band]

        # Get MTD XML file
        root, _ = self.read_mtd()

        # Convert DN to TOA radiance
        # <MEASURE_DESC>Raw radiometric counts (DN) to TOA Radiance (L). Formulae L=DN/GAIN+BIAS</MEASURE_DESC>
        try:
            rad_gain = None
            rad_bias = None
            for br in root.iterfind(".//Spectral_Band_Info"):
                if br.findtext("BAND_DESCRIPTION") == band_mtd_str:
                    rad_gain = float(br.findtext("PHYSICAL_GAIN"))
                    rad_bias = float(br.findtext("PHYSICAL_BIAS"))
                    break

            if rad_gain is None or rad_bias is None:
                raise TypeError

        except TypeError:
            raise InvalidProductError(
                "PHYSICAL_GAIN and PHYSICAL_BIAS from Spectral_Band_Info not found in metadata!"
            )
        return dn_arr / rad_gain + rad_bias

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = self.path / path.get_archived_path(
                    self.path, file_regex=r".*PREVIEW\.JPG"
                )
            else:
                quicklook_path = str(next(self.path.glob("*PREVIEW.JPG")))
        except (StopIteration, FileNotFoundError):
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    def _get_job_id(self) -> str:
        """
        Get VHR job ID

        Returns:
            str: VHR product ID
        """
        # Get MTD XML file
        root, _ = self.read_mtd()
        return root.findtext(".//JOB_ID")

    def _get_condensed_name(self) -> str:
        """
        Get PlanetScope products condensed name ({date}_PLD_{product_type}_{band_combi}).

        Returns:
            str: Condensed name
        """
        return f"{self.get_datetime()}_{self.constellation.name}_{self.product_type.name}_{self.band_combi.name}"



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" SAR Bands """
from eoreader.bands.band_names import SarBandNames
from eoreader.bands.bands import Band, BandMap
from eoreader.bands.mappings import EOREADER_TO_SPYNDEX_DICT
from eoreader.exceptions import InvalidTypeError


class SarBand(Band):
    """
    SAR Band object.
    Based on STAC band object.
    See `here <https://github.com/stac-extensions/eo/#band-object>`_ for more information, without useless information.
    """

    def __init__(self, eoreader_name, **kwargs):
        # Initialization from the super class
        super().__init__(eoreader_name, **kwargs)

        # Set names
        try:
            self.eoreader_name = SarBandNames.convert_from(eoreader_name)[0]
        except TypeError:
            raise InvalidTypeError

    def _to_repr_constellation_specific(self) -> list:
        """
        Representation specific to the constellation

        Returns:
            list: Representation list (constellation specific)
        """

        return []


# too many ancestors
# pylint: disable=R0901
class SarBandMap(BandMap):
    """SAR band map class"""

    def __init__(self) -> None:
        super().__init__({band_name: None for band_name in SarBandNames})

    def map_bands(self, band_map: dict) -> None:
        """
        Mapping band names to specific satellite band numbers, as strings.

        .. code-block:: python

            >>> sb = SarBandMap()
            >>> sb.map_bands({
                    VV: 1,
                })

        Args:
            band_map (dict): Band mapping as {SarBandNames: Band number for loading band}
        """
        for band_name, band in band_map.items():
            if not isinstance(band, SarBand):
                spyndex_name = (EOREADER_TO_SPYNDEX_DICT.get(band_name),)
                band = SarBand(
                    eoreader_name=band_name,
                    name=band,
                    id=band,
                    spyndex_name=spyndex_name,
                )

            if band_name not in self._band_map or not isinstance(
                band_name, SarBandNames
            ):
                raise InvalidTypeError(f"{band_name} should be a SarBandNames object")

            # Set number
            self._band_map[band_name] = band



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
COSMO-SkyMed products.
More info `here <https://egeos.my.salesforce.com/sfc/p/#1r000000qoOc/a/69000000JXxZ/WEEbowzi5cmY8vLqyfAAMKZ064iN1eWw_qZAgUkTtXI>`_.
"""
import logging
import os
import tempfile
from datetime import datetime
from enum import unique
from io import BytesIO
from typing import Union

import geopandas as gpd
import h5netcdf
import numpy as np
import rasterio
import xarray as xr
from lxml import etree
from lxml.builder import E
from rasterio import merge
from sertit import AnyPath, path, rasters, rasters_rio, strings, vectors
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType
from shapely.geometry import Polygon, box

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.exceptions import InvalidProductError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection
from eoreader.products.sar.sar_product import SAR_PREDICTOR

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class CosmoProductType(ListEnum):
    """
    COSMO-SkyMed (both generations) products types.

    The product classed are not specified here.

    More info
    `here <https://egeos.my.salesforce.com/sfc/p/#1r000000qoOc/a/69000000JXxZ/WEEbowzi5cmY8vLqyfAAMKZ064iN1eWw_qZAgUkTtXI>`_.
    """

    RAW = "RAW"
    """Level 0"""

    SCS = "SCS"
    """Level 1A, Single-look Complex Slant"""

    DGM = "DGM"
    """Level 1B, Detected Ground Multi-look"""

    GEC = "GEC"
    """Level 1C, Geocoded Ellipsoid Corrected"""

    GTC = "GTC"
    """Level 1D, Geocoded Terrain Corrected"""


class CosmoProduct(SarProduct):
    """
    Class for COSMO-SkyMed (both generations) Products
    More info
    `here <https://egeos.my.salesforce.com/sfc/p/#1r000000qoOc/a/69000000JXxZ/WEEbowzi5cmY8vLqyfAAMKZ064iN1eWw_qZAgUkTtXI>`_.
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        try:
            product_path = AnyPath(product_path)
            self._img_path = next(product_path.glob("*.h5"))
        except IndexError as ex:
            raise InvalidProductError(
                f"Image file (*.h5) not found in {product_path}"
            ) from ex

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # Private attributes
        self._raw_band_regex = "*_{}_*.h5"
        self._band_folder = self.path
        self.snap_filename = self._img_path.name

        # SNAP cannot process its archive
        self.needs_extraction = True

        # Get the number of swaths of this product
        with h5netcdf.File(self._img_path, phony_dims="access") as raw_h5:
            self.nof_swaths = len(list(raw_h5.groups))

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                                        geometry
            0  POLYGON ((108.09797 15.61011, 108.48224 15.678...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        root, _ = self.read_mtd()

        # Open extent coordinates
        try:

            def from_str_to_arr(geo_coord: str):
                return np.array(strings.str_to_list(geo_coord), dtype=float)[:2][::-1]

            bl_corner = from_str_to_arr(root.findtext(".//GeoCoordBottomLeft"))
            br_corner = from_str_to_arr(root.findtext(".//GeoCoordBottomRight"))
            tl_corner = from_str_to_arr(root.findtext(".//GeoCoordTopLeft"))
            tr_corner = from_str_to_arr(root.findtext(".//GeoCoordTopRight"))

            if bl_corner is None:
                raise InvalidProductError("Invalid XML: missing extent.")

            extent_wgs84 = gpd.GeoDataFrame(
                geometry=[Polygon([tl_corner, tr_corner, br_corner, bl_corner])],
                crs=vectors.WGS84,
            )
        except ValueError:

            def from_str_to_arr(geo_coord: str):
                str_list = [
                    it
                    for it in strings.str_to_list(geo_coord, additional_separator="\n")
                    if "+" not in it
                ]

                # Create tuples of 2D coords
                coord_list = []
                coord = [0.0, 0.0]
                for it_id, it in enumerate(str_list):
                    if it_id % 3 == 0:
                        # Invert lat and lon
                        coord[1] = float(it)
                    elif it_id % 3 == 1:
                        # Invert lat and lon
                        coord[0] = float(it)
                    elif it_id % 3 == 2:
                        # Z coordinates: do not store it

                        # Append the last coordinates
                        coord_list.append(coord.copy())

                        # And reinit it
                        coord = [0.0, 0.0]

                return coord_list

            bl_corners = from_str_to_arr(root.findtext(".//GeoCoordBottomLeft"))
            br_corners = from_str_to_arr(root.findtext(".//GeoCoordBottomRight"))
            tl_corners = from_str_to_arr(root.findtext(".//GeoCoordTopLeft"))
            tr_corners = from_str_to_arr(root.findtext(".//GeoCoordTopRight"))

            if not bl_corners:
                raise InvalidProductError("Invalid XML: missing extent.")

            assert (
                len(bl_corners) == len(br_corners) == len(tl_corners) == len(tr_corners)
            )

            polygons = [
                Polygon(
                    [
                        tl_corners[coord_id],
                        tr_corners[coord_id],
                        br_corners[coord_id],
                        bl_corners[coord_id],
                    ]
                )
                for coord_id in range(len(bl_corners))
            ]
            extents_wgs84 = gpd.GeoDataFrame(
                geometry=polygons,
                crs=vectors.WGS84,
            )

            extent_wgs84 = gpd.GeoDataFrame(
                geometry=[box(*extents_wgs84.total_bounds)],
                crs=vectors.WGS84,
            )

        return extent_wgs84

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, _ = self.read_mtd()

        # DGM_B, or SCS_B -> remove last 2 characters
        prod_type = root.findtext(".//ProductType")[:-2]
        if not prod_type:
            raise InvalidProductError("mode not found in metadata!")

        self.product_type = CosmoProductType.from_value(prod_type)

        if self.product_type == CosmoProductType.DGM:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type == CosmoProductType.SCS:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not available for {self.name}"
            )

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 10, 28, 22, 46, 25)
            >>> prod.get_datetime(as_datetime=False)
            '20201028T224625'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()

            # Open identifier
            acq_date = root.findtext(".//SceneSensingStartUTC")
            if not acq_date:
                raise InvalidProductError("SceneSensingStartUTC not found in metadata!")

            # Convert to datetime
            # 2020-10-28 22:46:24.808662850
            # To many milliseconds (strptime accepts max 6 digits) -> needs to be cropped
            date = datetime.strptime(acq_date[:-3], "%Y-%m-%d %H:%M:%S.%f")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        name = path.get_filename(root.findtext(".//ProductName"))
        if not name:
            raise InvalidProductError("ProductName not found in metadata!")

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1001513-735093"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element DeliveryNote at 0x2454ad4ee88>, {})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        try:
            mtd_from_path = "DFDN_*.h5.xml"

            return self._read_mtd_xml(mtd_from_path)
        except InvalidProductError:
            try:
                field_map = {
                    # ProductInfo
                    "ProductName": "Product Filename",
                    # "ProductId": ,
                    "MissionId": "Mission ID",
                    # "UniqueIdentifier": ,
                    "ProductGenerationDate": "Product Generation UTC",
                    # "UserRequestId": ,
                    # "ServiceRequestName": ,
                    # ProductDefinitionData
                    "ProductType": "Product Type",
                    "SceneSensingStartUTC": "Scene Sensing Start UTC",
                    "SceneSensingStopUTC": "Scene Sensing Stop UTC",
                    # "GeoCoordTopRightEN": ,
                    "GeoCoordSceneCentre": "Scene Centre Geodetic Coordinates",
                    "SatelliteId": "Satellite ID",
                    "AcquisitionMode": "Acquisition Mode",
                    "LookSide": "Look Side",
                    "ProjectionId": "Projection ID",
                    "DeliveryMode": "Delivery Mode",
                    "AcquisitionStationId": "Acquisition Station ID",
                    # ProcessingInfo
                    # "ProcessingLevel":,
                    # ProductCharacteristics
                    "AzimuthGeometricResolution": "Azimuth Geometric Resolution",
                    "GroundRangeGeometricResolution": "Ground Range Geometric Resolution",
                }

                sbi_field_map = {
                    "GeoCoordBottomLeft": "Bottom Left Geodetic Coordinates",
                    "GeoCoordBottomRight": "Bottom Right Geodetic Coordinates",
                    "GeoCoordTopLeft": "Top Left Geodetic Coordinates",
                    "GeoCoordTopRight": "Top Right Geodetic Coordinates",
                    # "GeoCoordTopRightEN": "Top Right East-North",
                    "NearLookAngle": "Near Look Angle",
                    "FarLookAngle": "Far Look Angle",
                }

                def h5_to_str(h5_val):
                    str_val = str(h5_val)
                    str_val = str_val.replace("[", "")
                    str_val = str_val.replace("]", "")
                    return str_val

                with h5netcdf.File(self._img_path) as netcdf_ds:
                    # Create XML attributes
                    global_attr = []
                    for xml_attr, h5_attr in field_map.items():
                        try:
                            global_attr.append(
                                E(xml_attr, h5_to_str(netcdf_ds.attrs[h5_attr]))
                            )
                        except KeyError:
                            # CSG products don't have their ProductName in the h5 file...
                            if xml_attr == "ProductName":
                                global_attr.append(
                                    E(xml_attr, path.get_filename(self._img_path))
                                )

                    try:
                        # CSK products
                        sbi = netcdf_ds.groups["S01"].variables["SBI"]
                    except KeyError:
                        # CSG products
                        sbi = netcdf_ds.groups["S01"].variables["IMG"]

                    for xml_attr, h5_attr in sbi_field_map.items():
                        global_attr.append(E(xml_attr, h5_to_str(sbi.attrs[h5_attr])))

                    mtd = E.s3_global_attributes(*global_attr)
                    mtd_el = etree.fromstring(
                        etree.tostring(
                            mtd,
                            pretty_print=True,
                            xml_declaration=True,
                            encoding="UTF-8",
                        )
                    )

                return mtd_el, {}
            except KeyError:
                raise InvalidProductError(
                    "Missing the XML metadata file. Cannot read the product."
                )

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        qlk_path, qlk_exists = self._get_out_path(f"{self.condensed_name}_QLK.tif")
        if not qlk_exists:
            with rasterio.open(str(self._img_path)) as ds:
                quicklook_paths = [subds for subds in ds.subdatasets if "QLK" in subds]

            if len(quicklook_paths) == 0:
                LOGGER.warning(f"No quicklook found in {self.condensed_name}")
            else:
                rasters.write(
                    rasters.read(quicklook_paths[0]),
                    qlk_path,
                    dtype=np.uint8,
                    nodata=255,
                )
                if len(quicklook_paths) > 1:
                    LOGGER.info(
                        "For now, only the quicklook of the first swath is taken into account."
                    )

        return str(qlk_path)

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        if path.is_cloud_path(self.path):
            h5_xarr_path = BytesIO(self._img_path.read_bytes())
        else:
            h5_xarr_path = str(self._img_path)

        with xr.open_dataset(
            h5_xarr_path, phony_dims="access", engine="h5netcdf"
        ) as h5_xarr:
            # Get the orbit direction
            try:
                od = OrbitDirection.from_value(getattr(h5_xarr, "Orbit Direction"))

            except TypeError:
                raise InvalidProductError("Orbit Direction not found in metadata!")

        return od

    def _pre_process_sar(self, band, pixel_size: float = None, **kwargs) -> str:
        """
        Pre-process SAR data (geocoding...)

        Args:
            band (sbn): Band to preprocess
            pixel_size (float): Pixl size
            kwargs: Additional arguments

        Returns:
            str: Band path
        """
        with h5netcdf.File(self._img_path, phony_dims="access") as raw_h5:
            if self.nof_swaths == 1:
                return super()._pre_process_sar(band, pixel_size, **kwargs)
            else:
                LOGGER.warning(
                    "Currently, SNAP doesn't handle multiswath Cosmo-SkyMed products. This is a workaround. See https://github.com/sertit/eoreader/issues/78"
                )

                # For every swath, pre-process the swath array alone
                pp_swath_path = []
                for group in raw_h5.groups:
                    with tempfile.TemporaryDirectory() as tmp_dir:
                        LOGGER.debug(f"Processing {group}")

                        # Create a mock-up of a COSMO product with only one swath and handled by SNAP
                        prod_path = os.path.join(
                            tmp_dir, f"{path.get_filename(self._img_path)}.h5"
                        )
                        with h5netcdf.File(
                            prod_path, "w", phony_dims="access"
                        ) as group_h5:
                            # Basic layer
                            group_h5.attrs.update(raw_h5.attrs)

                            # Change the swath to S01 as it is the only one read by SNAP (and is mandatory for the file to be recognized)
                            new_group = "S01"
                            group_h5.create_group(new_group)
                            group_h5.groups[new_group].attrs.update(
                                raw_h5.groups[group].attrs
                            )

                            # Copy all variables
                            for var_name in raw_h5.groups[group].variables:
                                var = raw_h5.groups[group].variables[var_name]
                                group_h5.groups[new_group].create_variable(
                                    f"/{new_group}/{var_name}",
                                    dimensions=var.dimensions,
                                    dtype=var.dtype,
                                    data=var,
                                    chunks=var.chunks,
                                )
                                group_h5.groups[new_group].variables[
                                    var_name
                                ].attrs.update(var.attrs)

                            # Copy all groups
                            for grp_name in raw_h5.groups[group].groups:
                                grp = raw_h5.groups[group].groups[grp_name]
                                if grp_name not in group_h5.groups[new_group].groups:
                                    group_h5.groups[new_group].create_group(grp_name)
                                    group_h5.groups[new_group].groups[
                                        grp_name
                                    ].attrs.update(grp.attrs)

                        # Pre-process swath
                        pp_swath_path.append(
                            super()._pre_process_sar(
                                band,
                                pixel_size,
                                prod_path=prod_path,
                                suffix=group,
                                **kwargs,
                            )
                        )

                # Merge the swaths
                LOGGER.debug("Merging the swaths")
                pp_path = os.path.join(
                    self._get_band_folder(writable=True),
                    f"{self.condensed_name}_{band.value.upper()}.tif",
                )
                # Force GTiff to be used in SNAP
                # Don't use rasters.merge_gtiff because off the predictor and the nodata...
                try:
                    pp_ds = [rasterio.open(p) for p in pp_swath_path]
                    merged_array, merged_transform = merge.merge(pp_ds, **kwargs)
                    merged_meta = pp_ds[0].meta.copy()
                    merged_meta.update(
                        {
                            "driver": "GTiff",
                            "height": merged_array.shape[1],
                            "width": merged_array.shape[2],
                            "transform": merged_transform,
                        }
                    )
                finally:
                    for ds in pp_ds:
                        ds.close()

                # Write
                # WARNING: Set nodata to 0 here as it is the value wanted by SNAP !

                # SNAP fails with classic predictor !!! Set the predictor to the default value (1) !!!
                # Caused by: javax.imageio.IIOException: Illegal value for Predictor in TIFF file
                # https://forum.step.esa.int/t/exception-found-when-reading-compressed-tif/654/7
                rasters_rio.write(
                    merged_array,
                    merged_meta,
                    pp_path,
                    nodata=self._snap_no_data,
                    predictor=SAR_PREDICTOR,
                )

                return pp_path



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
COSMO-SkyMed products.
More info `here <https://earth.esa.int/documents/10174/465595/COSMO-SkyMed-Mission-Products-Description>`_.
"""
import logging
from enum import unique

from sertit.misc import ListEnum

from eoreader import EOREADER_NAME
from eoreader.exceptions import InvalidProductError
from eoreader.products import CosmoProduct, CosmoProductType

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class CskSensorMode(ListEnum):
    """
    COSMO-SkyMed sensor mode.
    Take a look `here <https://earth.esa.int/documents/10174/465595/COSMO-SkyMed-Mission-Products-Description>`_.
    """

    HI = "HIMAGE"
    """Himage"""

    PP = "PINGPONG"
    """PingPong"""

    WR = "WIDEREGION"
    """Wide Region"""

    HR = "HUGEREGION"
    """Huge Region"""

    S2 = "ENHANCED SPOTLIGHT"
    """Enhanced Spotlight"""


class CskProduct(CosmoProduct):
    """
    Class for COSMO-SkyMed Products

    .. code-block:: python

        >>> from eoreader.reader import Reader
        >>> # CSK products could have any folder but needs to have a .h5 file correctly formatted
        >>> # ie. "CSKS1_SCS_B_HI_15_HH_RA_SF_20201028224625_20201028224632.h5"
        >>> path = r"1011117-766193"
        >>> prod = Reader().open(path)
    """

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://earth.esa.int/eogateway/documents/20142/37627/COSMO-SkyMed-Mission-Products-Description.pdf>`_
        for more information (p. 30)
        """
        if self.sensor_mode == CskSensorMode.HI:
            def_pixel_size = 2.5
            if self.product_type == CosmoProductType.SCS:
                def_res = 3.0
            else:
                def_res = 5.0
        elif self.sensor_mode == CskSensorMode.PP:
            def_pixel_size = 10.0
            def_res = 20.0
        elif self.sensor_mode == CskSensorMode.WR:
            def_pixel_size = 15.0
            def_res = 30.0
        elif self.sensor_mode == CskSensorMode.HR:
            def_pixel_size = 50.0
            def_res = 100.0
        elif self.sensor_mode == CskSensorMode.S2:
            def_pixel_size = 0.5
            def_res = 1.0
        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        self.pixel_size = def_pixel_size
        self.resolution = def_res

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """
        if self.nof_swaths > 1:
            # Calibration fails with CSG data
            LOGGER.debug(
                "SNAP Error: Calibration currently fails for CSK data with multiple swaths. Removing this step."
            )
            self._calibrate = False

        # Post init done by the super class
        super()._post_init(**kwargs)

    def _set_sensor_mode(self) -> None:
        """Get sensor mode"""
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        acq_mode = root.findtext(".//AcquisitionMode")
        if not acq_mode:
            raise InvalidProductError("AcquisitionMode not found in metadata!")

        # Get sensor mode
        self.sensor_mode = CskSensorMode.from_value(acq_mode)

        if not self.sensor_mode:
            raise InvalidProductError(
                f"Invalid {self.constellation.value} name: {self.name}"
            )

    def _set_instrument(self) -> None:
        """
        Set instrument

        CSK: https://earth.esa.int/eogateway/missions/cosmo-skymed
        """
        self.instrument = "SAR-2000"



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RapidEye products.
See
`Product specifications <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
and `Planet documentation <https://developers.planet.com/docs/data/rapideye/>`_
for more information.
"""
import logging
from datetime import datetime
from enum import unique
from typing import Union

import numpy as np
import xarray as xr
from sertit import files, path
from sertit.misc import ListEnum
from sertit.types import AnyPathType

from eoreader import DATETIME_FMT, EOREADER_NAME
from eoreader.bands import (
    BLUE,
    GREEN,
    NARROW_NIR,
    NIR,
    RED,
    VRE_1,
    BandNames,
    SpectralBand,
)
from eoreader.exceptions import InvalidProductError
from eoreader.products.optical.planet_product import PlanetProduct
from eoreader.stac import GSD, ID, NAME, WV_MAX, WV_MIN

LOGGER = logging.getLogger(EOREADER_NAME)

_RE_EAI = {
    BLUE: 1997.8,
    GREEN: 1863.5,
    RED: 1560.4,
    VRE_1: 1395.0,
    NIR: 1124.4,
    NARROW_NIR: 1124.4,
}
"""
From RE `Product specifications <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
"""


@unique
class ReProductType(ListEnum):
    """
    RapidEye product types (processing levels)

    See `Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
    for more information.
    """

    L1B = "RapidEye Basic Scene Product"
    """
    Radiometric and sensor corrections applied to the data.
    On-board spacecraft attitude and ephemeris applied to the data.
    """

    L3A = "RapidEye Ortho Tile Product"
    """
    Radiometric and sensor corrections applied to the data.
    Imagery is orthorectified using the RPCs and an elevation model.
    """


class ReProduct(PlanetProduct):
    """
    Class of PlanetScope products.
    See `Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
    for more information.

    The scaling factor to retrieve the calibrated radiance is 0.01.
    """

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self.constellation = self._get_constellation()
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        try:
            if self.is_archived:
                path.get_archived_path(self.path, r".*udm\.tif")
            else:
                next(self.path.glob("**/*udm.tif"))
            self._has_udm = True
        except (FileNotFoundError, StopIteration):
            # Some RE products don't have udm files
            LOGGER.warning(
                "UDM mask not found. This product won't be cleaned and won't have any cloud band."
            )
            pass

        self._has_cloud_cover = True

        # Pre init done by the super class
        super()._post_init(**kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        self.pixel_size = 5.0

    def _set_instrument(self) -> None:
        """
        Set instrument

        See: https://space.oscar.wmo.int/instruments/view/reis
        """
        self.instrument = "REIS"

    def _map_bands(self):
        """
        Map bands
        See <Product specs `https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf`>_ for more information.
        """
        gsd = 6.5

        blue = SpectralBand(
            eoreader_name=BLUE,
            **{NAME: "Blue", ID: 1, GSD: gsd, WV_MIN: 440, WV_MAX: 510},
        )

        green = SpectralBand(
            eoreader_name=GREEN,
            **{NAME: "Green", ID: 2, GSD: gsd, WV_MIN: 520, WV_MAX: 590},
        )

        red = SpectralBand(
            eoreader_name=RED,
            **{NAME: "Red", ID: 3, GSD: gsd, WV_MIN: 630, WV_MAX: 685},
        )

        vre = SpectralBand(
            eoreader_name=VRE_1,
            **{NAME: "Red Edge", ID: 4, GSD: gsd, WV_MIN: 690, WV_MAX: 730},
        )

        nir = SpectralBand(
            eoreader_name=NIR,
            **{NAME: "NIR", ID: 5, GSD: gsd, WV_MIN: 760, WV_MAX: 850},
        )

        # Set the band map
        self.bands.map_bands(
            {
                BLUE: blue,
                GREEN: green,
                RED: red,
                VRE_1: vre,
                NIR: nir,
                NARROW_NIR: nir,
            }
        )

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Manage product type
        prod_type = root.findtext(f".//{nsmap['eop']}productType")
        if not prod_type:
            raise InvalidProductError(
                "Cannot find the product type in the metadata file"
            )

        # Set correct product type
        self.product_type = getattr(ReProductType, prod_type)
        if self.product_type == ReProductType.L1B:
            # TODO: implement orthorectification for Planet products
            raise NotImplementedError(
                f"Basic Scene Product are not managed for {self.constellation.value} products:\n{self.path}"
            )

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2019, 6, 25, 10, 57, 28, 756000), fetched from metadata, so we have the ms
            >>> prod.get_datetime(as_datetime=False)
            '20190625T105728'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, nsmap = self.read_mtd()
            datetime_str = root.findtext(f".//{nsmap['eop']}acquisitionDate")
            if not datetime_str:
                raise InvalidProductError(
                    "Cannot find acquisitionDate in the metadata file."
                )

            # Convert to datetime
            datetime_str = datetime_str.split(".")[
                0
            ]  # Too many microseconds, remove them
            datetime_str = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S")

        else:
            datetime_str = self.datetime

        if not as_datetime:
            datetime_str = datetime_str.strftime(DATETIME_FMT)

        return datetime_str

    def _get_stack_path(self, as_list: bool = False) -> Union[str, list]:
        """
        Get Planet stack path(s)

        Args:
            as_list (bool): Get stack path as a list (useful if several subdatasets are present)

        Returns:
            Union[str, list]: Stack path(s)
        """
        if self._merged:
            stack_path, _ = self._get_out_path(f"{self.condensed_name}_analytic.vrt")
            if as_list:
                stack_path = [stack_path]
        else:
            stack_path = self._get_path(
                self.name, "tif", invalid_lookahead=["_udm", "_browse"], as_list=as_list
            )
            if not stack_path:
                stack_path = self._get_path(self.name, "TIF", as_list=as_list)

        return stack_path

    def _dn_to_toa_rad(self, dn_arr: xr.DataArray, band: BandNames) -> xr.DataArray:
        """
        Compute DN to TOA radiance

        See
        `Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
        for more information.

        Args:
            dn_arr (xr.DataArray): DN array
            band (BandNames): Band

        Returns:
            xr.DataArray: TOA Radiance array
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Open identifier
        rad_coef = None
        for band_mtd in root.iterfind(
            f".//{nsmap[self._nsmap_key]}bandSpecificMetadata"
        ):
            if (
                int(band_mtd.findtext(f".//{nsmap[self._nsmap_key]}bandNumber"))
                == self.bands[band].id
            ):
                rad_coef = float(
                    band_mtd.findtext(
                        f".//{nsmap[self._nsmap_key]}radiometricScaleFactor"
                    )
                )
                break

        if rad_coef is None:
            raise InvalidProductError(
                "Couldn't find any radiometricScaleFactor in the product metadata!"
            )

        # To reflectance
        return dn_arr * rad_coef

    def _toa_rad_to_toa_refl(
        self, rad_arr: xr.DataArray, band: BandNames
    ) -> xr.DataArray:
        """
        Compute TOA reflectance from TOA radiance

        See
        `Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
        for more information.

        WARNING: in this formula, d**2 = 1 / sqrt(dt) !

        Args:
            rad_arr (xr.DataArray): TOA Radiance array
            band (BandNames): Band

        Returns:
            xr.DataArray: TOA Reflectance array
        """

        # Compute the coefficient converting TOA radiance in TOA reflectance
        dt = self._sun_earth_distance_variation() ** 2
        _, sun_zen = self.get_mean_sun_angles()
        rad_sun_zen = np.deg2rad(sun_zen)
        eai = _RE_EAI[band]
        toa_refl_coeff = np.pi / (eai * dt * np.cos(rad_sun_zen))

        return rad_arr.copy(data=toa_refl_coeff * rad_arr)

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        # Convert DN into radiance
        band_arr = self._dn_to_toa_rad(band_arr, band)

        # Convert radiance into reflectance
        band_arr = self._toa_rad_to_toa_refl(band_arr, band)

        # To float32
        if band_arr.dtype != np.float32:
            band_arr = band_arr.astype(np.float32)

        return band_arr

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = path.get_archived_rio_path(
                    self.path, file_regex=r".*_browse\.tif"
                )
            else:
                quicklook_path = str(next(self.path.glob("**/*_browse.tif")))
        except (StopIteration, FileNotFoundError):
            pass

        return quicklook_path

    def _merge_subdatasets_mtd(self):
        """
        Merge subdataset, when several Planet products avec been ordered together
        Will create a reflectance (if possible) VRT, a UDM/UDM2 VRT and a merged metadata XML file.
        """
        LOGGER.warning(
            "_merge_subdatasets_mtd is not yet implemented (because of lack of tiled RapideEye product), only copying the first metadata file!"
        )

        mtd_file, mtd_exists = self._get_out_path(
            f"{self.condensed_name}_metadata.json"
        )
        if not mtd_exists:
            files.copy(self._get_path("metadata", "xml"), mtd_file)



import os

import setuptools

from eoreader.__meta__ import (
    __author__,
    __author_email__,
    __description__,
    __documentation__,
    __title__,
    __url__,
    __version__,
)

BASEDIR = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))
with open(os.path.join(BASEDIR, "README.md"), "r", encoding="utf8") as f:
    readme = f.read()

setuptools.setup(
    name=__title__,
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "lxml",
        "h5netcdf",
        "scipy",
        "rasterio>=1.3.0",
        "xarray>=0.18.0",
        "rioxarray>=0.10.0",
        "geopandas>=0.11.0",
        "sertit[full]>=1.30.0",
        "spyndex>=0.3.0",
        "pyresample",
        "zarr",
        "rtree",
        "validators",
        "methodtools",
        "dicttoxml",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_data={"": ["LICENSE", "NOTICE"], "eoreader.data": ["*.xml"]},
    include_package_data=True,
    python_requires=">=3.9",
    project_urls={
        "Bug Tracker": f"{__url__}/issues/",
        "Documentation": __documentation__,
        "Source Code": __url__,
    },
)



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Sentinel-3 SLSTR products

.. WARNING:
    Not georeferenced NetCDF files are badly opened by GDAL and therefore by rasterio !
    -> use xr.open_dataset that manages that correctly
"""
import logging
from collections import namedtuple
from enum import unique
from functools import reduce
from typing import Union

import numpy as np
import xarray as xr
from rasterio import features
from rasterio.enums import Resampling
from sertit import path, rasters
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME, cache, utils
from eoreader.bands import (
    ALL_CLOUDS,
    CIRRUS,
    CLOUDS,
    F1,
    F2,
    GREEN,
    NARROW_NIR,
    NIR,
    RAW_CLOUDS,
    RED,
    S7,
    SWIR_1,
    SWIR_2,
    SWIR_CIRRUS,
    TIR_1,
    TIR_2,
    BandNames,
    SpectralBand,
    to_str,
)
from eoreader.exceptions import InvalidTypeError
from eoreader.keywords import CLEAN_OPTICAL, SLSTR_RAD_ADJUST, SLSTR_STRIPE, SLSTR_VIEW
from eoreader.products import S3DataType, S3Product, S3ProductType
from eoreader.products.optical.optical_product import DEF_CLEAN_METHOD, CleanMethod
from eoreader.stac import ASSET_ROLE, BT, CENTER_WV, DESCRIPTION, FWHM, GSD, ID, NAME

LOGGER = logging.getLogger(EOREADER_NAME)

# FROM SNAP (only for radiance bands, not for brightness temperatures)
# https://github.com/senbox-org/s3tbx/blob/197c9a471002eb2ec1fbd54e9a31bfc963446645/s3tbx-rad2refl/src/main/java/org/esa/s3tbx/processor/rad2refl/Rad2ReflConstants.java#L141
# Not used for now
SLSTR_SOLAR_FLUXES_DEFAULT = {
    GREEN: 1837.39,
    RED: 1525.94,
    NIR: 956.17,
    NARROW_NIR: 956.17,
    SWIR_CIRRUS: 365.90,
    SWIR_1: 248.33,
    SWIR_2: 78.33,
}

# Link band names to their stripe
SLSTR_A_BANDS = ["S1", "S2", "S3"]
SLSTR_ABC_BANDS = ["S4", "S5", "S6"]
SLSTR_F_BANDS = ["F1"]
SLSTR_I_BANDS = ["S7", "S8", "S9", "F1", "F2"]

# Link band names to their physical quantity (radiance vs brightness temperature)
SLSTR_RAD_BANDS = SLSTR_A_BANDS + SLSTR_ABC_BANDS
SLSTR_BT_BANDS = SLSTR_I_BANDS

# Radiance adjustment
# Nadir and Oblique
FIELDS = [f"{rad}_n" for rad in SLSTR_A_BANDS + SLSTR_ABC_BANDS] + [
    f"{rad}_o" for rad in SLSTR_A_BANDS + SLSTR_ABC_BANDS
]

SlstrRadAdjustTuple = namedtuple(
    "SlstrRadAdjustTuple", FIELDS, defaults=(1.0,) * len(FIELDS)
)


@unique
class SlstrView(ListEnum):
    """
    Sentinel-3 SLSTR views: nadir view (n) and oblique view (o)

    Used in the context:
        - "an" and "ao" refer to the 500 m grid, stripe A, respectively for nadir view (n) and oblique view (o)
        - "bn" and "bo" refer to the 500 m grid, stripe B
        - "cn" and "co" refer to the 500 m grid, stripe C
        - "in" and "io" refer to the 1 km grid
        - "fn" and "fo" refer to the F1 channel 1 km grid
    """

    NADIR = "n"
    """Nadir view (n)"""

    OBLIQUE = "o"
    """Oblique view (o)"""


@unique
class SlstrStripe(ListEnum):
    """
    Sentinel-3 SLSTR stripes for 500m data: A and B

    Used in the context:
        - "an" and "ao" refer to the 500 m grid, stripe A, respectively for nadir view (n) and oblique view (o)
        - "bn" and "bo" refer to the 500 m grid, stripe B
        - "cn" and "co" refer to the 500 m grid, TDI
        - "in" and "io" refer to the 1 km grid
        - "fn" and "fo" refer to the F1 channel 1 km grid
    """

    A = "a"
    """Stripe A (a)"""

    B = "b"
    """Stripe B (b)"""

    TDI = "c"
    """TDI (c)"""

    I = "i"  # noqa
    """Not really a stripe, but refers to the 1 km grid"""

    F = "f"
    """Not really a stripe, but refers to the F1 channel 1 km grid"""


class SlstrRadAdjust(ListEnum):
    """
    SLSTR Radiance Adjustment dictionaries.

    Sentinel-3 SLSTR radiometry is not nominal, therefore a first-order radiometric correction is provided.
    """

    SNAP = SlstrRadAdjustTuple(
        # Nadir
        S5_n=1.12,
        S6_n=1.13,
        # Oblique
        S5_o=1.15,
        S6_o=1.14,
    )
    """
    SNAP Radiometric adjustment used in S3MPC Adjustment (optional in SNAP). Coefficients can be seen
    [here](https://github.com/senbox-org/s3tbx/blob/b10514e399f7a8a436002d2bacdb0c62be72f8f8/s3tbx-sentinel3-reader/src/main/java/org/esa/s3tbx/dataio/s3/slstr/SlstrLevel1ProductFactory.java#L72-L75)
    """

    S3_PN_SLSTR_L1_06 = SlstrRadAdjustTuple(
        # Nadir
        S5_n=1.12,
        S6_n=1.15,
        # Oblique
        S5_o=1.20,
        S6_o=1.26,
    )
    """
    Coefficients given in the
    [Sentinel-3 Product Notice 06](https://www-cdn.eumetsat.int/files/2020-04/pdf_s3a_pn_slstr_l1_06.pdf),
    edited the 07/11/2018 and reviewed the 19/11/2018
    """

    S3_PN_SLSTR_L1_07 = S3_PN_SLSTR_L1_06
    """
    Coefficients given in the
    [Sentinel-3 Product Notice 07](https://www-cdn.eumetsat.int/files/2020-06/pdf_s3a_pn_slstr_l1_07_1.1.pdf),
    edited the 15/01/2020 and reviewed the 09/06/2020, same as the Product Notice 06.
    """

    S3_PN_SLSTR_L1_08 = SlstrRadAdjustTuple(
        # Nadir
        S1_n=0.97,
        S2_n=0.98,
        S3_n=0.98,
        S5_n=1.11,
        S6_n=1.13,
        # Oblique
        S1_o=0.94,
        S2_o=0.95,
        S3_o=0.95,
        S5_o=1.04,
        S6_o=1.07,
    )
    """
    Coefficients given in the
    [Sentinel-3 Product Notice 08](https://www-cdn.eumetsat.int/files/2021-05/S3.PN-SLSTR-L1.08%20-%20i1r0%20-%20SLSTR%20L1%20PB%202.75-A%20and%201.53-B.pdf),
    edited the 18/05/2021.

    The default one.
    """

    NONE = SlstrRadAdjustTuple()
    """
    Coefficients set to one.
    """


class S3SlstrProduct(S3Product):
    """
    Class of Sentinel-3 SLSTR Products
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        """
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-3-slstr/level-1/observation-mode-desc
        Note that the name of each netCDF file provides information about its content.

        The suffix of each filename is associated with the selected grid:

        - "an" and "ao" refer to the 500 m grid, stripe A, respectively for nadir view (n) and oblique view (o)
        - "bn" and "bo" refer to the 500 m grid, stripe B
        - "in" and "io" refer to the 1 km grid
        - "fn" and "fo" refer to the F1 channel 1 km grid
        - "tx/n/o" refer to the tie-point grid for agnostic/nadir and oblique view
        """
        self._flags_file = None
        self._cloud_name = None
        self._exception_name = None

        # Default stripe (A) and view (NADIR)
        self._stripe = SlstrStripe.A
        self._view = SlstrView.NADIR
        self._rad_adjust = SlstrRadAdjust.S3_PN_SLSTR_L1_08

        # Brightness temperature
        self._bt_file = "{band}_BT_{suffix}.nc"
        self._bt_subds = "{band}_BT_{suffix}"

        super().__init__(
            product_path, archive_path, output_path, remove_tmp, **kwargs
        )  # Order is important here, gcps NEED to be after this

        self._F1_is_f = True
        try:
            self._get_raw_band_path(F1)
        except (FileNotFoundError, StopIteration):
            self._F1_is_f = False

    def _get_preprocessed_band_path(
        self,
        filename: str,
        suffix: str,
        pixel_size: Union[float, tuple, list] = None,
        writable: bool = True,
    ) -> AnyPathType:
        """
        Create the pre-processed band path

        Args:
            filename (str): Filename
            pixel_size (Union[float, tuple, list]): Pixel size of the wanted UTM band
            writable (bool): Do we need to write the pre-processed band ?

        Returns:
            AnyPathType: Pre-processed band path
        """
        res_str = self._pixel_size_to_str(pixel_size)
        if filename.endswith(suffix):
            pp_name = f"{self.condensed_name}_{filename}_{res_str}.tif"
        else:
            pp_name = f"{self.condensed_name}_{filename}_{suffix}_{res_str}.tif"

        return self._get_band_folder(writable=writable).joinpath(pp_name)

    def _set_preprocess_members(self):
        """Set pre-process members"""
        # Radiance bands
        self._radiance_file = "{band}_radiance_{suffix}.nc"
        self._radiance_subds = "{band}_radiance_{suffix}"

        # Geocoding
        self._geo_file = "geodetic_{suffix}.nc"
        self._lat_nc_name = "latitude_{suffix}"
        self._lon_nc_name = "longitude_{suffix}"
        self._alt_nc_name = "elevation_{suffix}"

        # Tie geocoding
        self._tie_geo_file = "geodetic_tx.nc"
        self._tie_lat_nc_name = "latitude_tx"
        self._tie_lon_nc_name = "longitude_tx"

        # Mean Sun angles
        self._geom_file = "geometry_t{view}.nc"
        self._saa_name = "solar_azimuth_t{view}"
        self._sza_name = "solar_zenith_t{view}"

        # Rad 2 Refl
        self._misc_file = "{band}_quality_{suffix}.nc"
        self._solar_flux_name = "{band}_solar_irradiance_{suffix}"

        # Clouds
        self._flags_file = "flags_{suffix}.nc"
        self._cloud_name = "cloud_{suffix}"

        # Other
        self._exception_name = "{band}_exception_{suffix}"

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        self.pixel_size = 500.0

    def _set_product_type(self) -> None:
        """Set products type"""
        # Product type
        if self.name[7] != "1":
            raise InvalidTypeError("Only L1 products are used for Sentinel-3 data.")

        self.product_type = S3ProductType.SLSTR_RBT
        self._data_type = S3DataType.RBT

    def _map_bands(self) -> None:
        """
        Map bands
        """
        # Bands
        bt_res = 1000.0
        slstr_bands = {
            GREEN: SpectralBand(
                eoreader_name=GREEN,
                **{
                    NAME: SLSTR_A_BANDS[0],
                    ID: SLSTR_A_BANDS[0],
                    GSD: self.pixel_size,
                    CENTER_WV: 554.27,
                    FWHM: 19.26,
                    DESCRIPTION: "Cloud screening, vegetation monitoring, aerosol",
                },
            ),
            RED: SpectralBand(
                eoreader_name=RED,
                **{
                    NAME: SLSTR_A_BANDS[1],
                    ID: SLSTR_A_BANDS[1],
                    GSD: self.pixel_size,
                    CENTER_WV: 659.47,
                    FWHM: 19.25,
                    DESCRIPTION: "NDVI, vegetation monitoring, aerosol",
                },
            ),
            NIR: SpectralBand(
                eoreader_name=NIR,
                **{
                    NAME: SLSTR_A_BANDS[2],
                    ID: SLSTR_A_BANDS[2],
                    GSD: self.pixel_size,
                    CENTER_WV: 868.00,
                    FWHM: 20.60,
                    DESCRIPTION: "NDVI, cloud flagging, pixel co-registration",
                },
            ),
            NARROW_NIR: SpectralBand(
                eoreader_name=NARROW_NIR,
                **{
                    NAME: SLSTR_A_BANDS[2],
                    ID: SLSTR_A_BANDS[2],
                    GSD: self.pixel_size,
                    CENTER_WV: 868.00,
                    FWHM: 20.60,
                    DESCRIPTION: "NDVI, cloud flagging, pixel co-registration",
                },
            ),
            SWIR_CIRRUS: SpectralBand(
                eoreader_name=SWIR_CIRRUS,
                **{
                    NAME: SLSTR_ABC_BANDS[0],
                    ID: SLSTR_ABC_BANDS[0],
                    GSD: self.pixel_size,
                    CENTER_WV: 1374.80,
                    FWHM: 20.80,
                    DESCRIPTION: "Cirrus detection over land",
                },
            ),
            SWIR_1: SpectralBand(
                eoreader_name=SWIR_1,
                **{
                    NAME: SLSTR_ABC_BANDS[1],
                    ID: SLSTR_ABC_BANDS[1],
                    GSD: self.pixel_size,
                    CENTER_WV: 1613.40,
                    FWHM: 60.68,
                    DESCRIPTION: "Cloud clearing, ice, snow, vegetation monitoring",
                },
            ),
            SWIR_2: SpectralBand(
                eoreader_name=SWIR_2,
                **{
                    NAME: SLSTR_ABC_BANDS[2],
                    ID: SLSTR_ABC_BANDS[2],
                    GSD: bt_res,
                    CENTER_WV: 2255.70,
                    FWHM: 50.15,
                    DESCRIPTION: "Vegetation state and cloud clearing",
                },
            ),
            S7: SpectralBand(
                eoreader_name=S7,
                **{
                    NAME: SLSTR_I_BANDS[0],
                    ID: SLSTR_I_BANDS[0],
                    GSD: bt_res,
                    CENTER_WV: 3742.00,
                    FWHM: 398.00,
                    DESCRIPTION: "SST, LST, Active fire, brightness temperature, 1km",
                    ASSET_ROLE: BT,
                },
            ),
            TIR_1: SpectralBand(
                eoreader_name=TIR_1,
                **{
                    NAME: SLSTR_I_BANDS[1],
                    ID: SLSTR_I_BANDS[1],
                    GSD: bt_res,
                    CENTER_WV: 10854.00,
                    FWHM: 776.00,
                    DESCRIPTION: "SST, LST, Active fire, brightness temperature, 1km",
                    ASSET_ROLE: BT,
                },
            ),
            TIR_2: SpectralBand(
                eoreader_name=TIR_2,
                **{
                    NAME: SLSTR_I_BANDS[2],
                    ID: SLSTR_I_BANDS[2],
                    GSD: bt_res,
                    CENTER_WV: 12022.50,
                    FWHM: 905.00,
                    DESCRIPTION: "SST, LST, brightness temperature, 1km",
                    ASSET_ROLE: BT,
                },
            ),
            F1: SpectralBand(
                eoreader_name=F1,
                **{
                    NAME: SLSTR_I_BANDS[3],
                    ID: SLSTR_I_BANDS[3],
                    GSD: bt_res,
                    CENTER_WV: 3742.00,
                    FWHM: 398.00,
                    DESCRIPTION: "Active fire, brightness temperature, 1km",
                    ASSET_ROLE: BT,
                },
            ),
            F2: SpectralBand(
                eoreader_name=F2,
                **{
                    NAME: SLSTR_I_BANDS[4],
                    ID: SLSTR_I_BANDS[4],
                    GSD: bt_res,
                    CENTER_WV: 10854.00,
                    FWHM: 776.00,
                    DESCRIPTION: "Active fire, brightness temperature, 1km",
                    ASSET_ROLE: BT,
                },
            ),
        }

        # Bands
        self.bands.map_bands(slstr_bands)

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the raw band paths.

        Args:
            kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        raw_band_paths = {}
        for band in self.get_existing_bands():
            raw_band_paths[band] = self._get_raw_band_path(band, **kwargs)

        return raw_band_paths

    def _get_raw_band_path(self, band: BandNames, **kwargs) -> AnyPathType:
        """
        Return the raw band path.

        Args:
            band (BandNames): Wanted band
            kwargs: Additional arguments

        Returns:
            AnyPathType: Raw path of queried band
        """
        band_id = self.bands[band].id

        # Get this band's suffix
        suffix = kwargs.get("suffix", self._get_suffix(band, **kwargs))

        # Get band filename and subdataset
        if band_id in SLSTR_RAD_BANDS:
            filename = self._replace(self._radiance_file, band=band, suffix=suffix)
        elif band_id in SLSTR_BT_BANDS:
            filename = self._replace(self._bt_file, band=band, suffix=suffix)
        else:
            filename = band

        if self.is_archived:
            raw_path = path.get_archived_path(self.path, f".*{filename}*")
        else:
            raw_path = next(self.path.glob(f"*{filename}*"))

        return raw_path

    def _preprocess(
        self,
        band: Union[BandNames, str],
        pixel_size: float = None,
        to_reflectance: bool = True,
        subdataset: str = None,
        **kwargs,
    ) -> AnyPathType:
        """
        Pre-process S3 SLSTR bands:
        - Geocode
        - Adjust radiance
        - Convert radiance to reflectance

        Args:
            band (Union[BandNames, str]): Band to preprocess (quality flags or others are accepted)
            pixel_size (float): Pixl size
            to_reflectance (bool): Convert band to reflectance
            subdataset (str): Subdataset
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing {band: path}
        """
        if isinstance(band, BandNames):
            band_id = self.bands[band].id
            band_str = band.name
        else:
            band_id = band
            band_str = band

        # Get this band's suffix
        suffix = kwargs.get("suffix", self._get_suffix(band, **kwargs))

        # Get band filename and subdataset
        pp_name = subdataset if subdataset else band_str
        if band_id in SLSTR_RAD_BANDS:
            if not subdataset:
                subdataset = self._replace(
                    self._radiance_subds, band=band, suffix=suffix
                )
            filename = self._replace(self._radiance_file, band=band, suffix=suffix)
        elif band_id in SLSTR_BT_BANDS:
            if not subdataset:
                subdataset = self._replace(self._bt_subds, band=band, suffix=suffix)
            filename = self._replace(self._bt_file, band=band, suffix=suffix)
        else:
            filename = band

        # Get the pre-processed path
        pp_path = self._get_preprocessed_band_path(
            pp_name, suffix=suffix, pixel_size=pixel_size, writable=False
        )

        if not pp_path.is_file():
            pp_path = self._get_preprocessed_band_path(
                pp_name, suffix=suffix, pixel_size=pixel_size, writable=True
            )

            # Get raw band
            band_arr = self._read_nc(
                filename, subdataset, dtype=kwargs.get("dtype", np.float32)
            )

            # Radiance preprocess (BT bands are given in BT !)
            if not kwargs.get("flags", False) and band_id in SLSTR_RAD_BANDS:
                # Adjust radiance if needed
                # Get the user's radiance adjustment if existing
                rad_adjust = kwargs.get(SLSTR_RAD_ADJUST, self._rad_adjust)
                try:
                    # Try to convert the rad_adjust to the correct enum
                    rad_adjust = SlstrRadAdjust.from_value(rad_adjust)
                except ValueError:
                    # Allow the user to pass a custom SlstrRadAdjustTuple
                    assert isinstance(rad_adjust, SlstrRadAdjustTuple)

                band_arr = self._radiance_adjustment(
                    band_arr, band, view=suffix[-1], rad_adjust=rad_adjust
                )

                # Convert radiance to reflectances if needed
                # Convert first pixel by pixel before reprojection !
                if to_reflectance:
                    LOGGER.debug(f"Converting {band_str} to reflectance")
                    band_arr = self._rad_2_refl(band_arr, band, suffix)

                    # Debug
                    # utils.write(
                    #     band_arr,
                    #     self._get_band_folder(writable=True).joinpath(
                    #         f"{self.condensed_name}_{band.name}_rad2refl.tif"
                    #     ),
                    # )

            # Geocode
            LOGGER.debug(f"Geocoding {pp_name}")
            kwargs.pop("suffix", None)
            pp_arr = self._geocode(
                band_arr, pixel_size=pixel_size, suffix=suffix, **kwargs
            )

            # Write on disk
            utils.write(pp_arr, pp_path)

        return pp_path

    def _get_suffix(self, band: Union[str, BandNames] = None, **kwargs) -> str:
        """
        Get the suffix according to the (given) stripe and view.
            - "an" and "ao" refer to the 500 m grid, stripe A, respectively for nadir view (n) and oblique view (o)
            - "bn" and "bo" refer to the 500 m grid, stripe B
            - "cn" and "co" refer to the 500 m grid, stripe C
            - "in" and "io" refer to the 1 km grid
            - "fn" and "fo" refer to the F1 channel 1 km grid

        Args:
            band (Union[BandNames, str]): Band from which to get the stripe
            kwargs: Other arguments

        Returns:
            str: Suffix (an, bn, cn, in, fn, ao, bo, co, io, fo)
        """
        # Get the view
        view = SlstrView.from_value(kwargs.get(SLSTR_VIEW, self._view))

        if band is not None:
            # Get the stripe
            if isinstance(band, BandNames):
                band = self.bands[band].id

            if band == F1.value:
                if self._F1_is_f:
                    stripe = SlstrStripe.F
                else:
                    stripe = SlstrStripe.I
            else:
                if band in SLSTR_I_BANDS:
                    stripe = SlstrStripe.I
                elif band in SLSTR_ABC_BANDS:
                    stripe = SlstrStripe.from_value(
                        kwargs.get(SLSTR_STRIPE, SlstrStripe.A)
                    )
                else:
                    stripe = SlstrStripe.A
        else:
            stripe = SlstrStripe.from_value(kwargs.get(SLSTR_STRIPE, SlstrStripe.A))

        # Return the prefix
        return f"{stripe.value}{view.value}"

    def _tie_to_img(self, tie_arr: np.ndarray, suffix: str) -> np.ndarray:
        """
        Convert an image sampled on the tie point grid (tx) to the wanted gris, given by the suffix

        Args:
            tie_arr (xr.Dataset): Image sampled on the tie point grid (tx)
            suffix: Suffix of the new grid

        Returns:
            np.ndarray: Array resampled to the wanted grid as a numpy array
        """
        # Load tie point grid
        tie_cart_file = "cartesian_tx.nc"
        tx_nc_name = "x_tx"
        ty_nc_name = "y_tx"

        # WARNING: RectBivariateSpline must have increasing values
        tx = np.squeeze(self._read_nc(tie_cart_file, tx_nc_name).data)[0, ::-1]
        ty = np.squeeze(self._read_nc(tie_cart_file, ty_nc_name).data)[:, 0]

        # Load fill image grid (cartesian)
        geo_file = f"cartesian_{suffix}.nc"
        fx_nc_name = f"x_{suffix}"
        fy_nc_name = f"y_{suffix}"

        fx = np.squeeze(self._read_nc(geo_file, fx_nc_name))
        fy = np.squeeze(self._read_nc(geo_file, fy_nc_name))

        # Interpolate via Spline (as extrapolation is possible and the grid is very sparse along the rows)
        # Import scipy here (long import)
        from scipy.interpolate import RectBivariateSpline

        # WARNING 2: Rasterio reads like [count, y, x] !
        no_nan_arr = np.nan_to_num(np.squeeze(tie_arr).data[:, ::-1])
        spline_interp = RectBivariateSpline(ty, tx, no_nan_arr)

        # Interpolate and set nodata back
        img_arr = spline_interp.ev(fy, fx)
        img_arr[img_arr == 0] = np.nan

        return img_arr

    # def _bt_2_rad(self, band_arr: xr.DataArray, band: BandNames = None) -> xr.DataArray:
    #     """
    #     Convert brightness temperature to radiance
    #
    #     The Level-1 brightness temperature measurements provided for the thermal channels (S7-S9, F1 and F2)
    #     can be converted to radiance by integrating the Planck function at the BT of interest multiplied over the
    #     spectral response of each band. The spectral response functions for SLSTR-A and SLSTR-B are available on
    #     the ESA Sentinel Online website (see Section 8.2.10)
    #     https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/instrument/measured-spectral-response-function-data
    #
    #     In https://sentinel.esa.int/documents/247904/4598085/Sentinel-3-SLSTR-Land-Handbook.pdf/bee342eb-40d4-9b31-babb-8bea2748264a
    #
    #     Args:
    #         band_arr (xr.DataArray): Band array
    #         band (BandNames): Spectral Band
    #
    #     Returns:
    #         dict: Dictionary containing {band: path}
    #     """
    #
    #     return band_arr

    def _rad_2_refl(
        self, band_arr: xr.DataArray, band: BandNames, suffix: str
    ) -> xr.DataArray:
        """
        Convert radiance to reflectance

        The visible and SWIR channels (S1-S6) provide measurements of top of atmosphere (ToA) radiances
        (mW/m2/sr/nm). These values can be converted to normalised reflectance for better comparison or
        merging of data with different sun angles as follows:
        reflectance = π* (ToA radiance / solar irradiance / COS(solar zenith angle))
        where the solar irradiance at ToA is given in the ‘quality’ dataset for the channel,
        and the solar zenith angle is given in the ‘geometry’ dataset.

        The solar irradiance contained in the quality dataset is derived from the solar spectrum
        of Thuillier et al. (2003) integrated over the measured SLSTR spectral responses
        and corrected for the earth-to-sun distance at the time of the measurement.

        In https://sentinel.esa.int/documents/247904/4598085/Sentinel-3-SLSTR-Land-Handbook.pdf/bee342eb-40d4-9b31-babb-8bea2748264a

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Spectral Band
            suffix (str): Band suffix

        Returns:
            dict: Dictionary containing {band: path}
        """
        rad_2_refl_path, rad_2_refl_exists = self._get_out_path(
            f"rad_2_refl_{band.name}_{suffix}.npy"
        )
        if not rad_2_refl_exists:
            # Open SZA array (resampled to band_arr size)
            sza = self._compute_sza_img_grid(suffix)

            # Open solar flux (resampled to band_arr size)
            e0 = self._compute_e0(band, suffix)

            # Compute rad_2_refl coeff
            rad_2_refl_coeff = (np.pi / e0 / np.cos(sza)).astype(np.float32)

            # Write on disk
            np.save(str(rad_2_refl_path), rad_2_refl_coeff)

        else:
            # Open rad_2_refl_coeff (resampled to band_arr size)
            if path.is_cloud_path(rad_2_refl_path):
                rad_2_refl_path = rad_2_refl_path.fspath

            rad_2_refl_coeff = np.load(str(rad_2_refl_path))

        return band_arr * rad_2_refl_coeff

    def _radiance_adjustment(
        self,
        band_arr: xr.DataArray,
        band: Union[str, BandNames],
        view: str,
        rad_adjust: SlstrRadAdjust = SlstrRadAdjust.S3_PN_SLSTR_L1_08,
    ) -> xr.DataArray:
        """
        Applying the radiance adjustment as recommended in the product notice:
        S3.PN-SLSTR-L1.08 (https://www-cdn.eumetsat.int/files/2021-05/S3.PN-SLSTR-L1.08%20-%20i1r0%20-%20SLSTR%20L1%20PB%202.75-A%20and%201.53-B.pdf):

        SLSTR-A/B: All solar channels (S1-S6) have been undergoing a vicarious calibration assessment to
        quantify their radiometric calibration adjustment. Recent analysis of vicarious calibration results
        over desert sites performed by RAL, CNES, Rayference and University of Arizona have determined
        new and consistent radiometric deviations wrt. common reference sensors (MERIS, MODIS)
        [S3MPC.RAL.TN.010]. Consequently, these have been used to provide a first-order radiometric
        corrections which are provided in the below tables with more detail at the following link
        [S3MPC.RAL.TN.020]. Current radiances in the L1B product remain uncorrected of these
        radiometric calibration adjustments. Hence, these multiplicative coefficients are strongly
        recommended to be used by all users.

        Nadir view
                      S1   S2   S3   S5   S6
        Correction  0.97 0.98 0.98 1.11 1.13
        Uncertainty 0.03 0.02 0.02 0.02 0.02

        Oblique view
                      S1   S2   S3   S5   S6
        Correction  0.94 0.95 0.95 1.04 1.07
        Uncertainty 0.05 0.03 0.03 0.03 0.05

        Args:
            band_arr (xr.DataArray): Band array
            band (Union[str, BandNames]): Optical Band
            view (str): View (n or o for Nadir and Oblique)
            rad_adjust (SlstrRadAdjust): Radiance Adjustment

        Returns:
            xr.DataArray: Adjusted band array
        """
        try:
            band_id = self.bands[band].id
            if band_id in SLSTR_RAD_BANDS:
                # Allow the tuple and the enum
                if isinstance(rad_adjust, SlstrRadAdjust):
                    rad_adjust_tuple = rad_adjust.value
                else:
                    rad_adjust_tuple = rad_adjust

                # Get the band coefficient and multiply the band
                rad_coeff = getattr(rad_adjust_tuple, f"{band_id}_{view}")
                band_arr *= rad_coeff
        except KeyError:
            # Not a band (ie Quality Flags) or Brightness temperature: no adjust needed
            pass

        return band_arr

    def _compute_sza_img_grid(self, suffix) -> np.ndarray:
        """
        Compute Sun Zenith Angle (in radian) resampled to the image grid (from the tie point grid)

        Args:
            suffix (str): Suffix
        Returns:
            np.ndarray: Resampled Sun Zenith Angle as a numpy array
        """
        sza_img_path, sza_exists = self._get_out_path(f"sza_{suffix}.npy")
        if not sza_exists:
            geom_file = self._replace(self._geom_file, view=suffix[-1])
            sza_name = self._replace(self._sza_name, view=suffix[-1])
            sza = self._read_nc(geom_file, sza_name)
            sza_rad = sza * np.pi / 180.0

            # From tie grid to image grid
            sza_img = self._tie_to_img(sza_rad, suffix)

            # Write on disk
            np.save(str(sza_img_path), sza_img)

        else:
            # Open sza_img (resampled to band_arr size)
            if path.is_cloud_path(sza_img_path):
                sza_img_path = sza_img_path.fspath

            sza_img = np.load(str(sza_img_path))

        return sza_img

    def _compute_e0(self, band: BandNames, suffix: str) -> np.ndarray:
        """
        Compute the solar spectral flux in mW / (m^2 * sr * nm)

        Args:
            band (BandNames): Optical Band
            suffix (str): Suffix

        Returns:
            np.ndarray: Solar Flux

        """
        misc = self._replace(self._misc_file, band=band, suffix=suffix)
        solar_flux_name = self._replace(self._solar_flux_name, band=band, suffix=suffix)

        e0 = self._read_nc(misc, solar_flux_name).data
        e0 = np.nanmean(e0)
        if np.isnan(e0):
            e0 = SLSTR_SOLAR_FLUXES_DEFAULT[band]

        return e0

    def _manage_invalid_pixels(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)

        ISP_absent pixel_absent not_decompressed no_signal saturation invalid_radiance no_parameters unfilled_pixel

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as a SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Open quality flags
        # NOT OPTIMIZED, MAYBE CHECK INVALID PIXELS ON NOT GEOCODED DATA
        suffix = self._get_suffix(band, **kwargs)
        qual_flags_path = self._preprocess(
            band,
            suffix=suffix,
            subdataset=self._replace(self._exception_name, band=band, suffix=suffix),
            pixel_size=band_arr.rio.resolution(),
            to_reflectance=False,
            flags=True,
            dtype=np.uint8,
        )

        # Open flag file
        qual_arr = utils.read(
            qual_flags_path,
            size=(band_arr.rio.width, band_arr.rio.height),
            resampling=Resampling.nearest,  # Nearest to keep the flags
            masked=False,
            as_type=np.uint8,
            **kwargs,
        )

        # Set no data for everything that caused an exception (3 and more)
        exception = np.where(qual_arr >= 3, self._mask_true, self._mask_false)

        # Get nodata mask
        no_data = np.where(np.isnan(band_arr.data), self._mask_true, self._mask_false)

        # Combine masks
        mask = no_data | exception

        # DO not set 0 to epsilons as they are a part of the
        return self._set_nodata_mask(band_arr, mask)

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        -> SLSTR does
        """
        if band in [
            RAW_CLOUDS,
            ALL_CLOUDS,
            CLOUDS,
            CIRRUS,
        ]:
            has_band = True
        else:
            has_band = False

        return has_band

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Read S3 SLSTR clouds from the flags file:cloud netcdf file.
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-3-slstr/level-1/cloud-identification

        bit_id  flag_masks (ushort)     flag_meanings
        ===     ===                     ===
        0       1US                     visible
        1       2US                     1.37_threshold
        2       4US                     1.6_small_histogram
        3       8US                     1.6_large_histogram
        4       16US                    2.25_small_histogram
        5       32US                    2.25_large_histogram
        6       64US                    11_spatial_coherence
        7       128US                   gross_cloud
        8       256US                   thin_cirrus
        9       512US                   medium_high
        10      1024US                  fog_low_stratus
        11      2048US                  11_12_view_difference
        12      4096US                  3.7_11_view_difference
        13      8192US                  thermal_histogram
        14      16384US                 spare
        15      32768US                 spare

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}

        if bands:
            all_ids = list(np.arange(0, 14))
            cir_id = 8
            cloud_ids = [cid for cid in all_ids if cid != cir_id]

            # Open path
            suffix = self._get_suffix(**kwargs)
            flags_file = self._replace(self._flags_file, suffix=suffix)
            cloud_name = self._replace(self._cloud_name, suffix=suffix)
            cloud_path = self._preprocess(
                flags_file,
                suffix=suffix,
                subdataset=cloud_name,
                pixel_size=pixel_size,
                to_reflectance=False,
            )

            # Open cloud file
            clouds_array = utils.read(
                cloud_path,
                pixel_size=pixel_size,
                size=size,
                resampling=Resampling.nearest,
                masked=False,
                as_type=np.uint16,
                **kwargs,
            )

            # Get nodata mask
            nodata = np.where(np.isnan(clouds_array), 1, 0)

            for band in bands:
                if band == ALL_CLOUDS:
                    cloud = self._create_mask(clouds_array, all_ids, nodata)
                elif band == CLOUDS:
                    cloud = self._create_mask(clouds_array, cloud_ids, nodata)
                elif band == CIRRUS:
                    cloud = self._create_mask(clouds_array, cir_id, nodata)
                elif band == RAW_CLOUDS:
                    cloud = clouds_array
                else:
                    raise InvalidTypeError(
                        f"Non existing cloud band for Sentinel-3 SLSTR: {band}"
                    )

                # Rename
                band_name = to_str(band)[0]

                # Multi bands -> do not change long name
                if band != RAW_CLOUDS:
                    cloud.attrs["long_name"] = band_name
                band_dict[band] = cloud.rename(band_name).astype(np.float32)

        return band_dict

    def _create_mask(
        self,
        bit_array: xr.DataArray,
        bit_ids: Union[int, list],
        nodata: np.ndarray,
    ) -> xr.DataArray:
        """
        Create a mask masked array (uint8) from a bit array, bit IDs and a nodata mask.

        Args:
            bit_array (xr.DataArray): Conditional array
            bit_ids (Union[int, list]): Bit IDs
            nodata (np.ndarray): Nodata mask

        Returns:
            xr.DataArray: Mask masked array

        """
        if not isinstance(bit_ids, list):
            bit_ids = [bit_ids]
        conds = rasters.read_bit_array(bit_array, bit_ids)
        cond = reduce(lambda x, y: x | y, conds)  # Use every condition (bitwise or)

        cond_arr = np.where(cond, self._mask_true, self._mask_false).astype(np.uint8)
        cond_arr = np.squeeze(cond_arr)
        try:
            cond_arr = features.sieve(cond_arr, size=10, connectivity=4)
        except TypeError:
            # Manage dask arrays that fails with rasterio sieve
            cond_arr = features.sieve(cond_arr.compute(), size=10, connectivity=4)
        cond_arr = np.expand_dims(cond_arr, axis=0)

        return super()._create_mask(bit_array, cond_arr, nodata)

    @cache
    def get_mean_sun_angles(self, view: SlstrView = SlstrView.NADIR) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = "S3B_SL_1_RBT____20191115T233722_20191115T234022_20191117T031722_0179_032_144_3420_LN2_O_NT_003.SEN3"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (78.55043955912154, 31.172127033319388)

        Args:
            view (SlstrView): SLSTR View (Nadir or Oblique)
        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        geom_file = self._replace(self._geom_file, view=view.value)
        saa_name = self._replace(self._saa_name, view=view.value)
        sza_name = self._replace(self._sza_name, view=view.value)

        # Open sun azimuth and zenith files
        sun_az = self._read_nc(geom_file, saa_name)
        sun_ze = self._read_nc(geom_file, sza_name)

        return float(sun_az.mean().data) % 360, float(sun_ze.mean().data)

    def _get_clean_band_path(
        self,
        band: BandNames,
        pixel_size: float = None,
        writable: bool = False,
        **kwargs,
    ) -> AnyPathType:
        """
        Get clean band path.

        The clean band is the opened band where invalid pixels have been managed.

        Args:
            band (BandNames): Wanted band
            pixel_size (float): Band pixel size in meters
            kwargs: Additional arguments

        Returns:
            AnyPathType: Clean band path
        """
        cleaning_method = CleanMethod.from_value(
            kwargs.get(CLEAN_OPTICAL, DEF_CLEAN_METHOD)
        )

        suffix = self._get_suffix(band, **kwargs)
        res_str = self._pixel_size_to_str(pixel_size)

        return self._get_band_folder(writable).joinpath(
            f"{self.condensed_name}_{band.name}_{suffix}_{res_str.replace('.', '-')}_{cleaning_method.value}.tif",
        )



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https:#sertit.unistra.fr/
# This file is part of eoreader project
#     https:#github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Sentinel-3 OLCI products

.. WARNING:
    Not georeferenced NetCDF files are badly opened by GDAL and therefore by rasterio !
    -> use xr.open_dataset that manages that correctly
"""
import logging
from typing import Union

import numpy as np
import rasterio
import xarray as xr
from rasterio.enums import Resampling
from sertit import path, rasters, rasters_rio
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME, cache, utils
from eoreader.bands import (
    BLUE,
    CA,
    GREEN,
    GREEN_1,
    NARROW_NIR,
    NIR,
    RED,
    VRE_1,
    VRE_2,
    VRE_3,
    WV,
    YELLOW,
    BandNames,
    Oa01,
    Oa02,
    Oa09,
    Oa10,
    Oa13,
    Oa14,
    Oa15,
    Oa18,
    Oa19,
    Oa21,
    SpectralBand,
)
from eoreader.exceptions import InvalidTypeError
from eoreader.products import S3DataType, S3Product, S3ProductType
from eoreader.stac import CENTER_WV, DESCRIPTION, FWHM, GSD, ID, NAME

LOGGER = logging.getLogger(EOREADER_NAME)

# FROM SNAP:
# https://github.com/senbox-org/s3tbx/blob/197c9a471002eb2ec1fbd54e9a31bfc963446645/s3tbx-rad2refl/src/main/java/org/esa/s3tbx/processor/rad2refl/Rad2ReflConstants.java#L97
# Not used for now
OLCI_SOLAR_FLUXES_DEFAULT = {
    Oa01: 1714.9084,
    Oa02: 1872.3961,
    CA: 1926.6102,
    BLUE: 1930.2483,
    GREEN_1: 1804.2762,
    GREEN: 1651.5836,
    YELLOW: 1531.4067,
    RED: 1475.615,
    Oa09: 1408.9949,
    Oa10: 1265.5425,
    VRE_1: 1255.4227,
    VRE_2: 1178.0286,
    Oa13: 955.07043,
    Oa14: 914.18945,
    Oa15: 882.8275,
    VRE_3: 882.8275,
    NIR: 882.8275,
    NARROW_NIR: 882.8275,
    Oa18: 882.8275,
    Oa19: 882.8275,
    WV: 882.8275,
    Oa21: 882.8275,
}


class S3OlciProduct(S3Product):
    """
    Class of Sentinel-3 OLCI Products
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        """
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-3-slstr/level-1/observation-mode-desc

        Note that the name of each netCDF file provides information about its content.
        """
        super().__init__(
            product_path, archive_path, output_path, remove_tmp, **kwargs
        )  # Order is important here, gcps NEED to be after this

    def _get_preprocessed_band_path(
        self,
        band: Union[BandNames, str],
        pixel_size: Union[float, tuple, list] = None,
        writable=True,
    ) -> AnyPathType:
        """
        Create the pre-processed band path

        Args:
            band (band: Union[BandNames, str]): Wanted band (quality flags accepted)
            pixel_size (Union[float, tuple, list]): Resolution of the wanted UTM band
            writable (bool): Do we need to write the pre-processed band ?

        Returns:
            AnyPathType: Pre-processed band path
        """
        res_str = self._pixel_size_to_str(pixel_size)
        band_str = band.name if isinstance(band, BandNames) else band

        return self._get_band_folder(writable=writable).joinpath(
            f"{self.condensed_name}_{band_str}_{res_str}.tif"
        )

    def _set_preprocess_members(self):
        """Set pre-process members"""
        # Radiance bands
        self._radiance_file = "{band}_radiance.nc"
        self._radiance_subds = "{band}_radiance"

        # Geocoding
        self._geo_file = "geo_coordinates.nc"
        self._lat_nc_name = "latitude"
        self._lon_nc_name = "longitude"
        self._alt_nc_name = "altitude"

        # Tie geocoding
        self._tie_geo_file = "tie_geo_coordinates.nc"
        self._tie_lat_nc_name = "latitude"
        self._tie_lon_nc_name = "longitude"

        # Mean Sun angles
        self._geom_file = "tie_geometries.nc"
        self._saa_name = "SAA"
        self._sza_name = "SZA"

        # Rad 2 Refl
        self._misc_file = "instrument_data.nc"
        self._solar_flux_name = "solar_flux"
        self._det_index = "detector_index"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self.needs_extraction = False

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        self.pixel_size = 300.0

    def _set_product_type(self) -> None:
        """
        Set products type

        More on spectral bands `here <https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-olci/resolutions/radiometric>`_.
        """
        # Product type
        if self.name[7] != "1":
            raise InvalidTypeError("Only L1 products are used for Sentinel-3 data.")

        self.product_type = S3ProductType.OLCI_EFR
        self._data_type = S3DataType.EFR

    def _map_bands(self) -> None:
        """
        Map bands
        """
        # Bands
        olci_bands = {
            Oa01: SpectralBand(
                eoreader_name=Oa01,
                **{
                    NAME: "Oa01",
                    ID: 1,
                    GSD: self.pixel_size,
                    CENTER_WV: 400,
                    FWHM: 15,
                    DESCRIPTION: "Aerosol correction, improved water constituent retrieval",
                },
            ),
            Oa02: SpectralBand(
                eoreader_name=Oa02,
                **{
                    NAME: "Oa02",
                    ID: 2,
                    GSD: self.pixel_size,
                    CENTER_WV: 412.5,
                    FWHM: 10,
                    DESCRIPTION: "Yellow substance and detrital pigments (turbidity)",
                },
            ),
            CA: SpectralBand(
                eoreader_name=CA,
                **{
                    NAME: "Oa03",
                    ID: 3,
                    GSD: self.pixel_size,
                    CENTER_WV: 442.5,
                    FWHM: 10,
                    DESCRIPTION: "Chlorophyll absorption maximum, biogeochemistry, vegetation",
                },
            ),
            BLUE: SpectralBand(
                eoreader_name=BLUE,
                **{
                    NAME: "Oa04",
                    ID: 4,
                    GSD: self.pixel_size,
                    CENTER_WV: 490,
                    FWHM: 10,
                    DESCRIPTION: "High Chlorophyll",
                },
            ),
            GREEN_1: SpectralBand(
                eoreader_name=GREEN_1,
                **{
                    NAME: "Oa05",
                    ID: 5,
                    GSD: self.pixel_size,
                    CENTER_WV: 510,
                    FWHM: 10,
                    DESCRIPTION: "Chlorophyll, sediment, turbidity, red tide",
                },
            ),
            GREEN: SpectralBand(
                eoreader_name=GREEN,
                **{
                    NAME: "Oa06",
                    ID: 6,
                    GSD: self.pixel_size,
                    CENTER_WV: 560,
                    FWHM: 10,
                    DESCRIPTION: "Chlorophyll reference (Chlorophyll minimum)",
                },
            ),
            YELLOW: SpectralBand(
                eoreader_name=YELLOW,
                **{
                    NAME: "Oa07",
                    ID: 7,
                    GSD: self.pixel_size,
                    CENTER_WV: 620,
                    FWHM: 10,
                    DESCRIPTION: "Sediment loading",
                },
            ),
            RED: SpectralBand(
                eoreader_name=RED,
                **{
                    NAME: "Oa08",
                    ID: 8,
                    GSD: self.pixel_size,
                    CENTER_WV: 665,
                    FWHM: 10,
                    DESCRIPTION: "Chlorophyll (2nd Chlorophyll absorption maximum), sediment, yellow substance / vegetation",
                },
            ),
            Oa09: SpectralBand(
                eoreader_name=Oa09,
                **{
                    NAME: "Oa09",
                    ID: 9,
                    GSD: self.pixel_size,
                    CENTER_WV: 673.75,
                    FWHM: 7.5,
                    DESCRIPTION: "For improved fluorescence retrieval and to better account for smile together with the bands 665 and 680 nm",
                },
            ),
            Oa10: SpectralBand(
                eoreader_name=Oa10,
                **{
                    NAME: "Oa10",
                    ID: 10,
                    GSD: self.pixel_size,
                    CENTER_WV: 681.25,
                    FWHM: 7.5,
                    DESCRIPTION: "Chlorophyll fluorescence peak, red edge",
                },
            ),
            VRE_1: SpectralBand(
                eoreader_name=VRE_1,
                **{
                    NAME: "Oa11",
                    ID: 11,
                    GSD: self.pixel_size,
                    CENTER_WV: 708.75,
                    FWHM: 10,
                    DESCRIPTION: "Chlorophyll fluorescence baseline, red edge transition",
                },
            ),
            VRE_2: SpectralBand(
                eoreader_name=VRE_2,
                **{
                    NAME: "Oa12",
                    ID: 12,
                    GSD: self.pixel_size,
                    CENTER_WV: 753.75,
                    FWHM: 7.5,
                    DESCRIPTION: "O2 absorption/clouds, vegetation",
                },
            ),
            Oa13: SpectralBand(
                eoreader_name=Oa13,
                **{
                    NAME: "Oa13",
                    ID: 13,
                    GSD: self.pixel_size,
                    CENTER_WV: 761.25,
                    FWHM: 2.5,
                    DESCRIPTION: "O2 absorption band/aerosol correction.",
                },
            ),
            Oa14: SpectralBand(
                eoreader_name=Oa14,
                **{
                    NAME: "Oa14",
                    ID: 14,
                    GSD: self.pixel_size,
                    CENTER_WV: 764.375,
                    FWHM: 3.75,
                    DESCRIPTION: "Atmospheric correction",
                },
            ),
            Oa15: SpectralBand(
                eoreader_name=Oa15,
                **{
                    NAME: "Oa15",
                    ID: 15,
                    GSD: self.pixel_size,
                    CENTER_WV: 767.5,
                    FWHM: 2.5,
                    DESCRIPTION: "O2A used for cloud top pressure, fluorescence over land",
                },
            ),
            VRE_3: SpectralBand(
                eoreader_name=VRE_3,
                **{
                    NAME: "Oa16",
                    ID: 16,
                    GSD: self.pixel_size,
                    CENTER_WV: 778.75,
                    FWHM: 15,
                    DESCRIPTION: "Atmos. corr./aerosol corr.",
                },
            ),
            NIR: SpectralBand(
                eoreader_name=NIR,
                **{
                    NAME: "Oa17",
                    ID: 17,
                    GSD: self.pixel_size,
                    CENTER_WV: 865,
                    FWHM: 20,
                    DESCRIPTION: "Atmospheric correction/aerosol correction, clouds, pixel co-registration",
                },
            ),
            NARROW_NIR: SpectralBand(
                eoreader_name=NARROW_NIR,
                **{
                    NAME: "Oa17",
                    ID: 17,
                    GSD: self.pixel_size,
                    CENTER_WV: 865,
                    FWHM: 20,
                    DESCRIPTION: "Atmospheric correction/aerosol correction, clouds, pixel co-registration",
                },
            ),
            Oa18: SpectralBand(
                eoreader_name=Oa18,
                **{
                    NAME: "Oa18",
                    ID: 18,
                    GSD: self.pixel_size,
                    CENTER_WV: 885,
                    FWHM: 10,
                    DESCRIPTION: "Water vapour absorption reference band. Common reference band with SLSTR instrument. Vegetation monitoring",
                },
            ),
            Oa19: SpectralBand(
                eoreader_name=Oa19,
                **{
                    NAME: "Oa19",
                    ID: 19,
                    GSD: self.pixel_size,
                    CENTER_WV: 900,
                    FWHM: 10,
                    DESCRIPTION: "Water vapour absorption/vegetation monitoring (maximum reflectance)",
                },
            ),
            WV: SpectralBand(
                eoreader_name=WV,
                **{
                    NAME: "Oa20",
                    ID: 20,
                    GSD: self.pixel_size,
                    CENTER_WV: 940,
                    FWHM: 20,
                    DESCRIPTION: "Water vapour absorption, Atmospheric correction/aerosol correction",
                },
            ),
            Oa21: SpectralBand(
                eoreader_name=Oa21,
                **{
                    NAME: "Oa21",
                    ID: 21,
                    GSD: self.pixel_size,
                    CENTER_WV: 1020,
                    FWHM: 40,
                    DESCRIPTION: "Atmospheric correction/aerosol correction",
                },
            ),
        }
        self.bands.map_bands(olci_bands)

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the raw band paths.

        Args:
            kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        raw_band_paths = {}
        for band in self.get_existing_bands():
            # Get band filename and subdataset
            filename = self._replace(self._radiance_file, band=self.bands[band].name)

            if self.is_archived:
                raw_path = path.get_archived_path(self.path, f".*{filename}")
            else:
                raw_path = next(self.path.glob(f"*{filename}"))

            raw_band_paths[band] = raw_path

        return raw_band_paths

    def _preprocess(
        self,
        band: Union[BandNames, str],
        pixel_size: float = None,
        to_reflectance: bool = True,
        subdataset: str = None,
        **kwargs,
    ) -> AnyPathType:
        """
        Pre-process S3 OLCI bands:
        - Convert radiance to reflectance
        - Geocode

        Args:
            band (Union[BandNames, str]): Band to preprocess (quality flags or others are accepted)
            pixel_size (float): Pixl size
            to_reflectance (bool): Convert band to reflectance
            subdataset (str): Subdataset
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing {band: path}
        """
        band_str = band if isinstance(band, str) else band.name

        pp_path = self._get_preprocessed_band_path(
            band, pixel_size=pixel_size, writable=False
        )

        if not pp_path.is_file():
            pp_path = self._get_preprocessed_band_path(
                band, pixel_size=pixel_size, writable=True
            )

            # Get band regex
            if isinstance(band, BandNames):
                band_name = self.bands[band].name
                if not subdataset:
                    subdataset = self._replace(self._radiance_subds, band=band_name)
                filename = self._replace(self._radiance_file, band=band_name)
            else:
                filename = band

            # Get raw band
            band_arr = self._read_nc(
                filename, subdataset, dtype=kwargs.get("dtype", np.float32)
            )

            # Convert radiance to reflectances if needed
            # Convert first pixel by pixel before reprojection !
            if to_reflectance:
                LOGGER.debug(f"Converting {band_str} to reflectance")
                band_arr = self._rad_2_refl(band_arr, band)

            # Geocode
            LOGGER.debug(f"Geocoding {band_str}")
            pp_arr = self._geocode(band_arr, pixel_size=pixel_size, **kwargs)

            # Write on disk
            utils.write(pp_arr, pp_path)

        return pp_path

    def _rad_2_refl(
        self, band_arr: xr.DataArray, band: BandNames = None
    ) -> xr.DataArray:
        """
        Convert radiance to reflectance

        The Level 1 product provides measurements of top of atmosphere (ToA) radiances (mW/m2/sr/nm). These
        values can be converted to normalised reflectance for better comparison or merging of data with different
        sun angles as follows:
        reflectance = π* (ToA radiance / solar irradiance / cos(solar zenith angle))
        where the solar irradiance at ToA is given in the ‘solar_flux’ dataset  (instrument_data.nc  file)  for  each
        detector  and  each  channel,  and  the  solar  zenith  angle  is  given  at  Tie  Points  in the ‘SZA’ dataset
        (tie_geometry.nc file). The appropriate instrument detector is given at each pixel by the ‘detector_index’
        dataset (instrument_data.nc file).

        In https://sentinel.esa.int/documents/247904/4598069/Sentinel-3-OLCI-Land-Handbook.pdf/455f8c88-520f-da18-d744-f5cda41d2d91

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Spectral Band (for SLSTR only)

        Returns:
            dict: Dictionary containing {band: path}
        """
        rad_2_refl_path, rad_2_refl_exists = self._get_out_path(
            f"rad_2_refl_{band.name}.npy"
        )
        if not rad_2_refl_exists:
            # Open SZA array (resampled to band_arr size)
            sza_path, sza_exists = self._get_out_path("sza.tif")

            # May have been created before (don't recreate it)
            if not sza_exists and not sza_path.is_file():
                sza_nc = self._read_nc(self._geom_file, self._sza_name)
                utils.write(sza_nc, sza_path)

            with rasterio.open(str(sza_path)) as ds_sza:
                # Values can be easily interpolated at pixels from Tie Points by linear interpolation using the
                # image column coordinate.
                sza, _ = rasters_rio.read(
                    ds_sza,
                    size=(band_arr.rio.width, band_arr.rio.height),
                    resampling=Resampling.bilinear,
                    masked=False,
                )
                sza_rad = sza.astype(np.float32) * np.pi / 180.0

            # Open solar flux (resampled to band_arr size)
            e0 = self._compute_e0(band)

            # Compute rad_2_refl coeff
            rad_2_refl_coeff = (np.pi / e0 / np.cos(sza_rad)).astype(np.float32)

            # Write on disk
            np.save(str(rad_2_refl_path), rad_2_refl_coeff)

        else:
            # Open rad_2_refl_coeff (resampled to band_arr size)
            rad_2_refl_coeff = np.load(str(rad_2_refl_path))

        return band_arr * rad_2_refl_coeff

    def _compute_e0(self, band: BandNames = None) -> np.ndarray:
        """
        Compute the solar spectral flux in mW / (m^2 * sr * nm)

        Args:
            band (BandNames): Spectral Band (for SLSTR only)

        Returns:
            np.ndarray: Solar Flux

        """
        # Do not convert to int here as we want to keep the nans
        det_idx = self._read_nc(self._misc_file, self._det_index).data
        e0_det = self._read_nc(self._misc_file, self._solar_flux_name).data

        # Get band slice and open corresponding e0 for the detectors
        band_slice = int(self.bands[band].name[-2:]) - 1
        e0_det = np.squeeze(e0_det[0, band_slice, :])

        # Create e0
        e0 = det_idx
        not_nan_idx = ~np.isnan(det_idx)
        e0[not_nan_idx] = e0_det[det_idx[not_nan_idx].astype(int)]

        return e0

    def _manage_invalid_pixels(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...) for OLCI data.
        See there:
        https:sentinel.esa.int/documents/247904/1872756/Sentinel-3-OLCI-Product-Data-Format-Specification-OLCI-Level-1

        QUALITY FLAGS (From end to start of the 32 bits):
        | Bit |  Flag               |
        |----|----------------------|
        | 0  |   saturated21        |
        | 1  |   saturated20        |
        | 2  |   saturated19        |
        | 3  |   saturated18        |
        | 4  |   saturated17        |
        | 5  |   saturated16        |
        | 6  |   saturated15        |
        | 7  |   saturated14        |
        | 8  |   saturated13        |
        | 9  |   saturated12        |
        | 10 |   saturated11        |
        | 11 |   saturated10        |
        | 12 |   saturated09        |
        | 13 |   saturated08        |
        | 14 |   saturated07        |
        | 15 |   saturated06        |
        | 16 |   saturated05        |
        | 17 |   saturated04        |
        | 18 |   saturated03        |
        | 19 |   saturated02        |
        | 20 |   saturated01        |
        | 21 |   dubious            |
        | 22 |   sun-glint_risk     |
        | 23 |   duplicated         |
        | 24 |   cosmetic           |
        | 25 |   invalid            |
        | 26 |   straylight_risk    |
        | 27 |   bright             |
        | 28 |   tidal_region       |
        | 29 |   fresh_inland_water |
        | 30 |   coastline          |
        | 31 |   land               |

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Bit ids
        band_bit_id = {
            Oa01: 20,  # Band 1
            Oa02: 19,  # Band 2
            CA: 18,  # Band 3
            BLUE: 17,  # Band 4
            GREEN_1: 16,  # Band 5
            GREEN: 15,  # Band 6
            YELLOW: 14,  # Band 7
            RED: 13,  # Band 8
            Oa09: 12,  # Band 9
            Oa10: 11,  # Band 10
            VRE_1: 10,  # Band 11
            VRE_2: 9,  # Band 12
            Oa13: 8,  # Band 13
            Oa14: 7,  # Band 14
            Oa15: 6,  # Band 15
            VRE_3: 5,  # Band 16
            NIR: 4,  # Band 17
            NARROW_NIR: 4,  # Band 17
            Oa18: 3,  # Band 18
            Oa19: 2,  # Band 19
            WV: 1,  # Band 20
            Oa21: 0,  # Band 21
        }
        invalid_id = 25
        sat_band_id = band_bit_id[band]

        # Open quality flags
        # NOT OPTIMIZED, MAYBE CHECK INVALID PIXELS ON NOT GEOCODED DATA
        qual_regex = "qualityFlags"
        subds = "quality_flags"
        qual_flags_path = self._preprocess(
            qual_regex,
            subdataset=subds,
            pixel_size=band_arr.rio.resolution(),
            to_reflectance=False,
        )

        # Open flag file
        qual_arr = utils.read(
            qual_flags_path,
            size=(band_arr.rio.width, band_arr.rio.height),
            resampling=Resampling.nearest,  # Nearest to keep the flags
            masked=False,
            as_type=np.uint32,
            **kwargs,
        )
        invalid, sat = rasters.read_bit_array(qual_arr, [invalid_id, sat_band_id])

        # Get nodata mask
        no_data = np.where(np.isnan(band_arr.data), self._mask_true, self._mask_false)

        # Combine masks
        mask = no_data | invalid | sat

        return self._set_nodata_mask(band_arr, mask)

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        -> OLCI does not provide any cloud mask
        """
        return False

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Does nothing for OLCI data

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        if bands:
            LOGGER.warning("Sentinel-3 OLCI L1B does not provide any cloud file")
        return {}

    @cache
    def get_mean_sun_angles(self) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = "S3B_SL_1_RBT____20191115T233722_20191115T234022_20191117T031722_0179_032_144_3420_LN2_O_NT_003.SEN3"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (78.55043955912154, 31.172127033319388)

        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        # Open sun azimuth and zenith files
        sun_az = self._read_nc(self._geom_file, self._saa_name)
        sun_ze = self._read_nc(self._geom_file, self._sza_name)

        return float(sun_az.mean().data) % 360, float(sun_ze.mean().data)
